#!/usr/bin/env python3
"""
This module contains all the code used to test the testee module.
"""


import unittest
from os import path

from auxi.core.bibliography import Database as testee


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class DatabaseTester(unittest.TestCase):
    """
    The function tester for the bibliography Database class.
    """

    def test_construct(self):
        """
        Test whether an exception is raised when a database object is created
        from a bibtex file.
        """

        testee(path.join(path.dirname(path.realpath(__file__)),
               'data/bibliography.bib'))

    def test_entry(self):
        """
        Test whether an entry is successfully retrieved from the database.
        """

        db = testee(path.join(path.dirname(path.realpath(__file__)),
                    'data/bibliography.bib'))
        entry = db['lienhard2015']
        self.assertEqual(entry['author'],
                         'J.H. Lienhard IV and J.H. Lienhard V')
        self.assertEqual(entry['title'],
                         'A Heat Transfer Textbook')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
A toolkit to help metallurgical process engineers to rapidly do day-to-day
calculations.

auxi is a toolkit to help metallurgical process engineers with their day-to-day
tasks. Many of the calculations that we do require things like molar masses,
conversion of one compound to another using stoichiometry, heat transfer
calculations, mass balances, energy balances, etc. It is usually quite time
consuming to get started with these calculations in a tool like Excel. auxi
aims to save you time by making many of these calculations available from
within python.

We hope that auxi will help you spend less time focusing on searching for
formulas and data, and setting up calculations, and more on thinking about the
problems that you need to solve with these calculations. Enjoy!

If you need more information about how to use auxi, please refer the
documentation at https://auxi.readthedocs.org/en/latest/.
"""


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


#COMMON ISSUES
#TODO: Make sure that you update version numbers to include release candidate
#      numbers when releasing.
#TODO: Make sure that you stick to the 80 character column limit.
#TODO: Use standard docstring layout. Triple quotation marks on own lines.
#TODO: Use proper grammar and spelling in all comments and documentation.
#TODO: Add the following as first line to all code files and test:
#      #!/usr/bin/env python3
#TODO: Add authorship information according to:
#      http://stackoverflow.com/questions/1523427/what-is-the-common-header-format-of-python-files
#TODO: Remove '# -*- coding: utf-8 -*-' from top of code files.
#TODO: When running code file, run corresponding unit tests.
#TODO: Use three regions of imports: builtin, third party and own.
#TODO: Explictly state the purpose of each unit test in a docstring.



#!/usr/bin/env python3
"""
This module provides classes to create reports.
"""

from io import StringIO

from enum import Enum
import csv
from tabulate import tabulate

from auxi.core.objects import Object

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ReportFormat(Enum):
    """
    Represents the format the report should be outputted as.
    """

    printout = 1,
    latex = 2,
    txt = 3,
    csv = 4,
    string = 5,
    matplotlib = 6,
    png = 7


class Report(Object):
    """
    Base class for all auxi reports.
    """

    def __init__(self, data_source, output_path=None):
        self.data_source = data_source
        self.output_path = output_path

    def _generate_table_(self):
        return []

    def _render_matplotlib_(self, png=False):
        pass

    def render(self, format=ReportFormat.printout):
        """
        Render the report in the specified format

        :param format: The format. The default format is to print
          the report to the console.

        :returns: If the format was set to 'string' then a string
          representation of the report is returned.
        """

        table = self._generate_table_()
        if(format == ReportFormat.printout):
            print(tabulate(table, headers="firstrow", tablefmt="simple"))
        elif (format == ReportFormat.latex):
            self._render_latex_(table)
        elif (format == ReportFormat.txt):
            self._render_txt_(table)
        elif (format == ReportFormat.csv):
            self._render_csv_(table)
        elif (format == ReportFormat.string):
            return str(tabulate(table, headers="firstrow", tablefmt="simple"))
        elif (format == ReportFormat.matplotlib):
            self._render_matplotlib_()
            pass
        elif (format == ReportFormat.png):
            if self.output_path is None:
                self._render_matplotlib_()
            else:
                self._render_matplotlib_(True)
            pass

    def _render_latex_(self, table):
        if self.output_path is not None:
            with open(self.output_path + '.tex', 'w') as f:
                f.write(
                    tabulate(table, headers="firstrow", tablefmt="latex"))
        else:
            print(tabulate(table, headers="firstrow", tablefmt="latex"))

    def _render_txt_(self, table):
        if self.output_path is not None:
            with open(self.output_path + '.txt', 'w') as f:
                f.write(
                    tabulate(table, headers="firstrow", tablefmt="simple"))
        else:
            print(tabulate(table, headers="firstrow", tablefmt="simple"))

    def _render_csv_(self, table):
        if self.output_path is not None:
            with open(self.output_path + '.csv', 'w') as f:
                csv.writer(f, lineterminator='\n').writerows(table)
        else:
            with StringIO() as f:
                csv.writer(f, lineterminator='\n').writerows(table)
                print(f.getvalue())


if __name__ == "__main__":
    import unittest
    from reporting_test import ReportingUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides classes to manage currencies.
"""

from datetime import datetime

from enum import Enum

from auxi.core.helpers import get_date
from auxi.core.reporting import Report as ReportBase

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ReportType(Enum):
    """
    Represents a report type, e.g. balance sheet or income statement.
    """

    balance_sheet = 1,
    income_statement = 2,
    transaction_list = 3,
    cash_flow = 4


class Report(ReportBase):
    """
    Base class for reports.

    :param data_source: The object to report on.
    :param output_path: The path to write the report file to.
    """

    def __init__(self, data_source, output_path=None):
        super(Report, self).__init__(data_source, output_path)


class GeneralLedgerStructure(Report):
    """
    Report on a general ledger structure.

    :param data_source: The object to report on.
    :param output_path: The path to write the report file to.
    """

    def __init__(self, data_source, output_path=None):
        super(GeneralLedgerStructure, self).__init__(data_source, output_path)

    def _gl_accounts_to_table_(self, table, account, prefix=''):
        table.append([account.account_type.name,
                      prefix + account.number,
                      account.name, account.description])
        for acc in sorted(account.accounts, key=lambda a: a.number):
            self._gl_accounts_to_table_(
                table, acc, account.number + '/')

    def _generate_table_(self):
        table = [["Type", "Number", "Name", "Description"]]

        i = [a for a in self.data_source.accounts if a.number.startswith('I')]
        for acc in sorted(i, key=lambda a: a.number):
            self._gl_accounts_to_table_(table, acc)

        b = [a for a in self.data_source.accounts if a.number.startswith('B')]
        for acc in sorted(b, key=lambda a: a.number):
            self._gl_accounts_to_table_(table, acc)
        return table


class TransactionList(Report):
    """
    Report on a list of transactions.

    :param data_source: The object to report on.
    :param end: The start date to generate the report for.
    :param end: The end date to generate the report for.
    :param component_path: The path of the component to filter the report's
      transactions by.
    :param output_path: The path to write the report file to.
    """

    # TODO: Generate transaction for a component at a path only.

    def __init__(self, data_source, start=datetime.min.date(),
                 end=datetime.max.date(), component_path="",
                 output_path=None):
        self.start_date = get_date(start)
        self.end_date = get_date(end)
        self.component_path = component_path
        super(TransactionList, self).__init__(data_source, output_path)

    def _generate_table_(self):
        table = [["Date", "Source", "Tx Name", "Debit Account",
                  "Credit Account", "Amount", "Description"]]
        amount_tot = 0
        tx_list = self.data_source.transactions
        for t in sorted(tx_list, key=lambda v: v.tx_date):
            if not t.source.startswith(self.component_path):
                continue
            if t.tx_date >= self.start_date and \
               t.tx_date <= self.end_date:
                if t.source.endswith(t.name):
                    name = ''
                else:
                    name = t.name
                table.append([t.tx_date, t.source, name,
                              t.dt_account, t.cr_account,
                              "%.2f" % t.amount,
                              "" if t.description is None else t.description])
                amount_tot += t.amount
        table.append(["", "", "Total", "", "", "%.2f" % amount_tot, ""])
        return table


class BalanceSheet(Report):
    """
    Report a balance sheet of a general ledger.

    :param data_source: The object to report on.
    :param end: The end date to generate the report for.
    :param output_path: The path to write the report file to.
    """
    # TODO: Generate transaction for a component at a path only.

    def __init__(self, data_source, end=datetime.max.date(), output_path=None):
        """
        """
        self.end_date = get_date(end)
        super(BalanceSheet, self).__init__(data_source, output_path)

    def _sum_amounts_per_account_(self):
        gls = self.data_source.structure
        summedAssets = {}
        summedLiability = {}
        summedEquity = {}

        for t in self.data_source.transactions:
            if t.tx_date <= self.end_date:
                cr_account_type = gls.get_account(t.cr_account).account_type
                dt_account_type = gls.get_account(t.dt_account).account_type
                if cr_account_type == AccountType.asset:
                    if t.cr_account in summedAssets:
                        summedAssets[t.cr_account] -= t.amount
                    else:
                        summedAssets[t.cr_account] = -t.amount
                elif cr_account_type == AccountType.liability:
                    if t.cr_account in summedLiability:
                        summedLiability[t.cr_account] += t.amount
                    else:
                        summedLiability[t.cr_account] = t.amount
                elif cr_account_type == AccountType.equity:
                    if t.cr_account in summedEquity:
                        summedEquity[t.cr_account] += t.amount
                    else:
                        summedEquity[t.cr_account] = t.amount
                if dt_account_type == AccountType.asset:
                    if t.dt_account in summedAssets:
                        summedAssets[t.dt_account] += t.amount
                    else:
                        summedAssets[t.dt_account] = t.amount
                elif dt_account_type == AccountType.liability:
                    if t.dt_account in summedLiability:
                        summedLiability[t.dt_account] -= t.amount
                    else:
                        summedLiability[t.dt_account] = -t.amount
                elif dt_account_type == AccountType.equity:
                    if t.dt_account in summedEquity:
                        summedEquity[t.dt_account] -= t.amount
                    else:
                        summedEquity[t.dt_account] = -t.amount

        return summedAssets, summedLiability, summedEquity

    def _generate_table_(self):
        table = [["Assets", " ", "Liabilities and Equity", "", "  "]]

        sumAssets, sumLiability, sumEquity = self._sum_amounts_per_account_()

        assets_count = len(sumAssets)
        lia_n_eq_count = len(sumLiability) + len(sumEquity) + 4 + 3

        assets_sum = sum(sumAssets.values())
        liabilities_sum = sum(sumLiability.values())
        equities_sum = sum(sumEquity.values())

        rows = []
        for i in range(0, max(assets_count, lia_n_eq_count)):
            rows.append(["", "", "", "", ""])
        ix = 0
        for entry in sumAssets:
            rows[ix][0] = entry
            rows[ix][1] = "%.2f" % sumAssets[entry]
            ix += 1

        rows[0][2] = "Liabilities"
        rows[1][2] = "-----------"
        ix = 2
        for entry in sumLiability:
            rows[ix][2] = entry
            rows[ix][3] = "%.2f" % sumLiability[entry]
            ix += 1

        rows[ix][2] = "Total liabilities"
        rows[ix][4] = "%.2f" % liabilities_sum
        ix += 1
        rows[ix][2] = "-----------------"
        ix += 1
        rows[ix][2] = "Owners' Equity"
        ix += 1
        rows[ix][2] = "--------------"
        ix += 1
        for entry in sumEquity:
            rows[ix][2] = entry
            rows[ix][3] = "%.2f" % sumEquity[entry]
            ix += 1
        rows[ix][2] = "Total owners' equities"
        rows[ix][4] = "%.2f" % sum(sumEquity.values())

        rows.append(["-----", "--------", "----------------------", "",
                     "--------"])
        rows.append(["Total", "%.2f" % assets_sum, "Total", "",
                     "%.2f" % (liabilities_sum + equities_sum)])

        [table.append(row) for row in rows]

        return table


class IncomeStatement(Report):
    """
    Report an income statement of a general ledger.

    :param data_source: The object to report on.
    :param end: The start date to generate the report for.
    :param end: The end date to generate the report for.
    :param component_path: The path of the component to filter the report's
      transactions by.
    :param output_path: The path to write the report file to.
    """
    # TODO: Generate transaction for a component at a path only.

    def __init__(self, data_source, start=datetime.min.date(),
                 end=datetime.max.date(), component_path="", output_path=None):
        """
        """
        self.start_date = get_date(start)
        self.end_date = get_date(end)
        self.component_path = component_path
        super(IncomeStatement, self).__init__(data_source, output_path)

    def _sum_amounts_per_account_(self):
        gls = self.data_source.structure
        summedIncome = {}
        summedExpenses = {}
        for t in self.data_source.transactions:
            if not t.source.startswith(self.component_path):
                continue
            if t.tx_date >= self.start_date and t.tx_date <= self.end_date:
                cr_account_type = gls.get_account(t.cr_account).account_type
                dt_account_type = gls.get_account(t.dt_account).account_type
                if not t.is_closing_cr_account and \
                   not t.is_closing_dt_account and t.cr_account != "":
                    if cr_account_type == AccountType.revenue:
                        if t.cr_account in summedIncome:
                            summedIncome[t.cr_account] += t.amount
                        else:
                            summedIncome[t.cr_account] = t.amount
                    elif cr_account_type == AccountType.expense:
                        if t.cr_account in summedExpenses:
                            summedExpenses[t.cr_account] -= t.amount
                        else:
                            summedExpenses[t.cr_account] = t.amount
                if not t.is_closing_cr_account and \
                   not t.is_closing_dt_account and t.dt_account != "":
                    if dt_account_type == AccountType.revenue:
                        if t.dt_account in summedIncome:
                            summedIncome[t.dt_account] -= t.amount
                        else:
                            summedIncome[t.dt_account] = t.amount
                    elif dt_account_type == AccountType.expense:
                        if t.dt_account in summedExpenses:
                            summedExpenses[t.dt_account] += t.amount
                        else:
                            summedExpenses[t.dt_account] = t.amount
        return summedIncome, summedExpenses

    def _generate_table_(self):
        summedIncome, summedExpenses = self._sum_amounts_per_account_()

        sum_incomes = sum(summedIncome.values())
        sum_expenses = sum(summedExpenses.values())
        net_income = sum_incomes - sum_expenses

        table = [["", "Debit", "Credit"]]
        table.append(["Revenues", "", ""])
        table.append(["--------", "", ""])
        for entry in summedIncome:
            table.append([entry, "", "%.2f" % summedIncome[entry]])
        table.append(["", "", ""])
        table.append(["Cost of Sales", "", ""])
        table.append(["-------------", "", ""])
        cos_sum = 0
        for entry in summedExpenses:
            if entry.startswith('Cost of Sales'):
                cos_sum += summedExpenses[entry]
                table.append([entry, "%.2f" % summedExpenses[entry], ""])
        table.append(["", "", "--------"])
        table.append(["Gross Revenues (including interest income)",
                     "", "%.2f" % (sum_incomes - cos_sum)])
        table.append(["", "", "--------"])

        table.append(["Expenses", "", ""])
        table.append(["--------", "", ""])
        for entry in summedExpenses:
            if not entry.startswith('Cost of Sales'):
                table.append([entry, "%.2f" % summedExpenses[entry], ""])
        table.append(["", "--------", ""])
        table.append(["Total Expenses", "%.2f" % (sum_expenses - cos_sum), ""])
        table.append(["", "--------", "--------"])
        if net_income < 0:
            table.append(["Net Income", "", "(%.2f)" % abs(net_income)])
        else:
            table.append(["Net Income", "", "%.2f" % abs(net_income)])
        return table


from auxi.modelling.financial.des import AccountType


if __name__ == "__main__":
    import unittest
    from auxi.modelling.reporting import GeneralLedgerStructureUnitTester
    from auxi.modelling.reporting import TransactionListUnitTester
    from auxi.modelling.reporting import BalanceSheetUnitTester
    from auxi.modelling.reporting import IncomeStatementUnitTester
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides testing code for the thermochemistry module.
"""

import unittest

from auxi.tools.chemistry import thermochemistry as thermo


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


# TODO: Create tests for: write_compound_to_auxi_file, load_data_factsage,
#   load_data_auxi, list_compounds, molar_mass
# TODO: Test CpRecord, Phase and Compound classes sperately?

class ThermoFunctionTester(unittest.TestCase):
    """
    The function tester for the thermochemistry module.
    """

    def assertAlmostEqual(self, first, second, places=14, msg=None,
                          delta=None):
        if type(first) is list and type(second) is list:
            self.assertEqual(len(first), len(second))
            for f, s in zip(first, second):
                self.assertAlmostEqual(f, s)
        else:
            super(ThermoFunctionTester, self).assertAlmostEqual(
                first, second, places, msg, delta)

    def test_load_data_auxi(self):
        thermo.compounds.clear()
        thermo.load_data_auxi()
        self.assertEqual(len(thermo.compounds), 81)

    def test_compound_get_phase_list(self):
        phs = thermo.compounds["Ag"].get_phase_list()
        self.assertEqual(phs[0], "L")
        self.assertEqual(phs[1], "S")

    def test_phase___str__(self):
        phs = thermo.compounds["Ag"].get_phase_list()
        self.assertTrue(len(phs[0].__str__()) > 0)

    def test_cpRecord___str__(self):
        ph = thermo.compounds["Ag"]._phases['L']
        for t, cpr in ph._Cp_records.items():
            self.assertTrue(len(cpr.__str__()) > 0)
            break

    def test_get_datafile_references(self):
        refs = thermo.get_datafile_references()
        self.assertTrue("rao1985" in refs)
        self.assertTrue(len(refs["rao1985"]) > 0)

    def test_get_reference(self):
        self.assertEqual("rao1985", thermo.compounds["Ag"].reference)
        self.assertTrue(len(thermo.compounds["Ag"].get_reference()) > 0)

    def test_molar_mass(self):
        self.assertAlmostEqual(thermo.molar_mass("FeO"), 0.0718444)

    def test_Cp(self):
        self.assertAlmostEqual(thermo.Cp("Al2O3[S]", 1000.0),
                               0.00034731792422193833)
        #                       0.00036892922607564924)

    def test_H(self):
        self.assertAlmostEqual(thermo.H("Al2O3[S]", 1000.0),
                               -4.264869218634823)
        #                       -4.192397399191783)

    def test_S(self):
        self.assertAlmostEqual(thermo.S("Al2O3[S]", 1000.0),
                               0.0005662425810664761)
        #                       0.0005987306605872277)

    def test_G(self):
        self.assertAlmostEqual(thermo.G("Al2O3[S]", 1000.0),
                               -4.985780960719607)
        #                       0.0005662425810664761)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides testing code for the auxi.modelling.financial.des module.
"""

import unittest
from datetime import datetime

from auxi.modelling.financial.des import AccountType
from auxi.modelling.financial.des import GeneralLedgerAccount
from auxi.modelling.financial.des import Transaction
from auxi.modelling.financial.des import TransactionTemplate
from auxi.modelling.financial.des import GeneralLedgerStructure
from auxi.modelling.financial.des import GeneralLedger
from auxi.core.reporting import ReportFormat

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class GeneralLedgerAccountUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.financial.des.GeneralLedgerAccount class.
    """

    def setUp(self):
        self.object = GeneralLedgerAccount("NameA",
                                           description="DescriptionA",
                                           number="010",
                                           account_type=AccountType.asset)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.number, "010")
        self.assertEqual(self.object.account_type, AccountType.asset)

    def test_set_parent_path(self):
        self.object.set_parent_path("accyA/accyB")
        self.assertEqual(self.object.path, "accyA/accyB/NameA")

    def test_create_account(self):
        new_account = self.object.create_account("TestA",
                                                 description="TestA_Desc",
                                                 number="011")
        self.assertEqual(new_account.name, "TestA")
        self.assertEqual(new_account.description, "TestA_Desc")
        self.assertEqual(new_account.number, "011")
        self.assertEqual(new_account.account_type, self.object.account_type)

        self.assertEqual(new_account, self.object.accounts[0])

    def test_remove_account(self):
        num_accounts = len(self.object.accounts)
        self.object.create_account("TestA",
                                   description="TestA_Desc",
                                   number="011")
        self.object.remove_account("TestA")
        self.assertEqual(len(self.object.accounts), num_accounts)

    def test_get_child_account(self):
        sub_acc = self.object.create_account(
            "TestA",
            description="TestA_Desc",
            number="011")
        sub_sub_acc = sub_acc.create_account(
            "TestA1",
            description="TestA1_Desc",
            number="012")
        result = self.object.get_child_account("TestA/TestA1")

        self.assertEqual(result.name, sub_sub_acc.name)
        self.assertEqual(result.description, sub_sub_acc.description)
        self.assertEqual(result.number, sub_sub_acc.number)
        self.assertEqual(result.account_type, sub_sub_acc.account_type)

        self.assertEqual(result, sub_sub_acc)


class TransactionUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.financial.des.Transaction class.
    """

    def setUp(self):
        self.object = Transaction("NameA",
                                  description="DescriptionA",
                                  tx_date=datetime(2016, 2, 1).date(),
                                  dt_account="Bank",
                                  cr_account="Sales",
                                  source="PigeonSales",
                                  amount=100.00)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.dt_account, "Bank")
        self.assertEqual(self.object.tx_date, datetime(2016, 2, 1).date())
        self.assertEqual(self.object.cr_account, "Sales")
        self.assertEqual(self.object.source, "PigeonSales")
        self.assertEqual(self.object.is_closing_cr_account, False)
        self.assertEqual(self.object.is_closing_dt_account, False)
        self.assertEqual(self.object.amount, 100.0)


class TransactionTemplateUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.financial.des.TransactionTemplate class.
    """

    def setUp(self):
        self.object = TransactionTemplate("NameA",
                                          description="DescriptionA",
                                          dt_account="Bank",
                                          cr_account="Sales")

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.dt_account, "Bank")
        self.assertEqual(self.object.cr_account, "Sales")


class GeneralLedgerStructureUnitTester(unittest.TestCase):
    """
      Tester for the auxi.modelling.financial.des.generalledgerstructure class.
    """

    def setUp(self):
        self.object = GeneralLedgerStructure("NameA",
                                             description="DescriptionA")

    def test_constructor(self):
        """
        Test the the variables has been initialised and that the default
        accounts has been created.
        """
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertNotEqual(self.object["Bank"], None)
        self.assertEqual(
            self.object['Unallocated Income Statement'].account_type,
            AccountType.revenue)
        self.assertEqual(self.object["Sales"].number, "I10")
        self.assertEqual(
            self.object["Sales"].account_type,
            AccountType.revenue)
        self.assertEqual(self.object["Cost of Sales"].number, "I15")
        self.assertEqual(
            self.object["Cost of Sales"].account_type,
            AccountType.expense)
        self.assertEqual(self.object["Other Income"].number, "I20")
        self.assertEqual(
            self.object["Other Income"].account_type,
            AccountType.revenue)
        self.assertEqual(self.object["Expense"].number, "I25")
        self.assertEqual(
            self.object["Expense"].account_type,
            AccountType.expense)
        self.assertEqual(self.object["Tax"].number, "I30")
        self.assertEqual(
            self.object["Tax"].account_type,
            AccountType.expense)
        self.assertEqual(self.object["Dividends"].number, "I35")
        self.assertEqual(
            self.object["Dividends"].account_type,
            AccountType.expense)
        self.assertEqual(self.object["Share Capital"].number, "B10")
        self.assertEqual(
            self.object["Share Capital"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Retained Income"].number, "B15")
        self.assertEqual(
            self.object["Retained Income"].account_type,
            AccountType.equity)
        self.assertEqual(self.object["Shareholders Loan"].number, "B20")
        self.assertEqual(
            self.object["Shareholders Loan"].account_type,
            AccountType.liability)
        self.assertEqual(self.object["Long Term Borrowing"].number, "B25")
        self.assertEqual(
            self.object["Long Term Borrowing"].account_type,
            AccountType.liability)
        self.assertEqual(
            self.object["Other Long Term Liabilities"].number,
            "B30")
        self.assertEqual(
            self.object["Other Long Term Liabilities"].account_type,
            AccountType.liability)
        self.assertEqual(self.object["Fixed Assets"].number, "B35")
        self.assertEqual(
            self.object["Fixed Assets"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Investments"].number, "B40")
        self.assertEqual(
            self.object["Investments"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Other Fixed Assets"].number, "B45")
        self.assertEqual(
            self.object["Other Fixed Assets"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Inventory"].number, "B50")
        self.assertEqual(
            self.object["Inventory"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Accounts Receivable"].number, "B55")
        self.assertEqual(
            self.object["Accounts Receivable"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Bank"].number, "B60")
        self.assertEqual(
            self.object["Bank"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Bank"]["Default"].number, "0000")
        self.assertEqual(
            self.object["Bank"]["Default"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Other Current Assets"].number, "B65")
        self.assertEqual(
            self.object["Other Current Assets"].account_type,
            AccountType.asset)
        self.assertEqual(self.object["Account Payable"].number, "B70")
        self.assertEqual(
            self.object["Account Payable"].account_type,
            AccountType.liability)
        self.assertEqual(self.object["Taxation"].number, "B75")
        self.assertEqual(
            self.object["Taxation"].account_type,
            AccountType.liability)
        self.assertEqual(
            self.object["Other Current Liabilities"].number,
            "B80")
        self.assertEqual(
            self.object["Other Current Liabilities"].account_type,
            AccountType.liability)

        self.assertEqual(
            self.object['Unallocated Income Statement']["Gross Profit"].number,
            "0000")
        self.assertEqual(
            self.object['Unallocated Income Statement']
            ["Gross Profit"].account_type,
            AccountType.revenue)
        self.assertEqual(
            self.object['Unallocated Income Statement']
            ["Income Summary"].number,
            "0010")
        self.assertEqual(
            self.object['Unallocated Income Statement']
            ["Income Summary"].account_type,
            AccountType.revenue)
        self.assertEqual(
            self.object['Retained Income']
            ["Retained Earnings"].number,
            "0000")
        self.assertEqual(
            self.object['Retained Income']
            ["Retained Earnings"].account_type,
            AccountType.equity)
        self.assertEqual(
            self.object['Tax']
            ["Income Tax Expense"].number,
            "0000")
        self.assertEqual(
            self.object['Tax']
            ["Income Tax Expense"].account_type,
            AccountType.expense)

        self.assertEqual(self.object.tax_payment_account, "Bank/Default")

    def test_get_account(self):
        self.object["Retained Income"].create_account(
            "TestA",
            number="010")
        acc = self.object["Bank"].create_account(
            "TestB",
            number="020")
        sub_acc = acc.create_account(
            "TestB1",
            description="TestB1_Desc",
            number="010")
        sub_acc.create_account(
            "TestB1.1",
            description="TestB1.1_Desc",
            number="010")
        orig = sub_acc.create_account(
            "TestB1.2",
            description="TestB1.1_Desc",
            number="011")
        result = self.object.get_account("Bank/TestB/TestB1/TestB1.2")

        self.assertEqual(result.name, orig.name)
        self.assertEqual(result.description, orig.description)
        self.assertEqual(result.number, orig.number)

    def test_get_account_decendants(self):
        # Set up this test.
        self.sales_fish_acc = self.object["Sales"].create_account(
            "SalesFish",
            description="Sales of Fish",
            number="010")
        self.sales_barracuda_acc = self.sales_fish_acc.create_account(
            "SalesBarracuda",
            description="Sales of Barracudas",
            number="010")
        self.sales_nemo_acc = self.sales_fish_acc.create_account(
            "SalesNemo",
            description="Sales of Nemos",
            number="020")
        # perform the test.
        result = self.object.get_account_decendants(self.object["Sales"])
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], self.sales_fish_acc)
        self.assertEqual(result[1], self.sales_barracuda_acc)
        self.assertEqual(result[2], self.sales_nemo_acc)

    def test_validate_account_name_valid(self):
        self.object.validate_account_names(
            ["Bank/Default", "Retained Income/Retained Earnings"])

    def test_validate_account_name_invalid(self):
        self.assertRaises(
            ValueError,
            self.object.validate_account_names,
            ["invalid_acc_name_a", "invalid_acc_name_b"])

    def test_report(self):
        report = self.object.report(ReportFormat.string)
        line1 = report.split("\n")[0]
        self.assertEqual(line1.replace(" ", ""), "TypeNumberNameDescription")


class GeneralLedgerUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.financial.des.GeneralLedger class.
    """

    def setUp(self):
        self.structure = GeneralLedgerStructure("NameA",
                                                description="DescriptionA")
        self.structure["Retained Income"].create_account(
            "TestA",
            description="TestA_Desc",
            number="010")
        self.object = GeneralLedger("NameA",
                                    self.structure,
                                    description="DescriptionA")

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.structure, self.object.structure)

    def test_create_transaction(self):
        new_tx = self.object.create_transaction(
            "TestA",
            description="TestA_Desc",
            tx_date=datetime(2016, 2, 1).date(),
            dt_account="Bank",
            cr_account="Sales",
            source="Peanut Sales",
            amount=20.00)

        tx_list = self.object.transactions
        self.assertEqual(new_tx.name, tx_list[0].name)
        self.assertEqual(new_tx.tx_date, tx_list[0].tx_date)
        self.assertEqual(new_tx.dt_account, tx_list[0].dt_account)
        self.assertEqual(new_tx.cr_account, tx_list[0].cr_account)
        self.assertEqual(new_tx.source, tx_list[0].source)
        self.assertEqual(new_tx.amount, tx_list[0].amount)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This package provides auxi's bibliography, and tools to interact with it.
"""


import bibtexparser
from os import path


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Database(object):
    """
    A bibliography database that contains and provides entries in a bibtex
    file.

    :param path: the path to the bibtex file to load into the database
    """

    def __init__(self, path):
        with open(path) as file:
            bibtex_str = file.read()

        bib_database = bibtexparser.loads(bibtex_str)
        self._dict = bib_database.get_entry_dict()

    def __str__(self):
        return str(self._dict)

    def __getitem__(self, key):
        return self._dict[key]

db = Database(path.join(path.dirname(path.realpath(__file__)),
              r'data/bibliography.bib'))


if __name__ == "__main__":
    import unittest
    from bibliography_test import DatabaseTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides tools for calculating material physical properties.
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import webbrowser
from cerberus import Validator, ValidationError

from auxi.tools.materialphysicalproperties.core import Model


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Marno Grewar, Christoff Kok, Johan Zietsman'
__credits__ = ['Marno Grewar', 'Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Marno Grewar'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


def _formula_string(coeffs, variable="x"):
    """
    Create a string representation of a polynomial formula.
    """
    template0 = "%.3E"
    template1 = "%+.3E%s"
    templaten = "%+.3E%s^%.0f"

    result = ""
    for power, coeff in enumerate(reversed(coeffs)):
        if power == 0:
            result += template0 % coeff
        elif power == 1:
            result += " " + template1 % (coeff, variable)
        else:
            result += " " + templaten % (coeff, variable, power)
    return "$%s$" % result


class PolynomialModelT(Model):
    """
    A model that uses polynomial fits on experimental data to describe the
    variation of a specific material physical property as a function of
    temperature.

    :param material: the name of the material being described, e.g. "Air"
    :param proprty: the name of the property being described, e.g. "density"
    :param symbol: the symbol of the property being described, e.g. "rho"
    :param display symbol: the display symbol of the property being described,\
    e.g. "\rho"
    :param units: the units used to express the property, e.g. "kg/m3"
    :param references: a list of literature references on which this model\
    implementation is based, e.g. ['lienhard2015', 'smith2006']
    :param datasets: a list of data sets on which this model implementation is\
    based, e.g. ['dataset-air-lienhard2015.csv']
    :param coeffs: polynomial coefficients sorted from highest to lowest power
    """

    def create(dataset, symbol, degree):
        """
        Create a model object from the data set for the property specified by
        the supplied symbol, using the specified polynomial degree.

        :param dataset: a DataSet object
        :param symbol: the symbol of the property to be described, e.g. 'rho'
        :param degree: the polynomial degree to use

        :returns: a new PolynomialModelT object
        """

        x_vals = dataset.data['T'].tolist()
        y_vals = dataset.data[symbol].tolist()
        coeffs = np.polyfit(x_vals, y_vals, degree)

        result = PolynomialModelT(dataset.material,
                                  dataset.names_dict[symbol],
                                  symbol, dataset.display_symbols_dict[symbol],
                                  dataset.units_dict[symbol],
                                  None, [dataset.name], coeffs)

        result.state_schema['T']['min'] = float(min(x_vals))
        result.state_schema['T']['max'] = float(max(x_vals))

        return result

    def __init__(self, material, proprty, symbol, display_symbol, units,
                 references, datasets, coeffs):
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0}}
        super().__init__(material, proprty, symbol, display_symbol, units,
                         state_schema, references, datasets)
        self._coeffs = [float(x) for x in coeffs]

    def calculate(self, **state):
        """
        Calculate the material physical property at the specified temperature
        in the units specified by the object's 'property_units' property.

        :param T: [K] temperature

        :returns: physical property value
        """
        super().calculate(**state)
        return np.polyval(self._coeffs, state['T'])

    def plot(self, dataset, path, show=False):
        with PdfPages(path) as pdf:
            x_vals = dataset.data['T'].tolist()
            y_vals = dataset.data[self.symbol].tolist()
            plt.plot(x_vals, y_vals, 'ro', alpha=0.4, markersize=4)

            x_vals2 = np.linspace(min(x_vals), max(x_vals), 80)
            fx = np.polyval(self._coeffs, x_vals2)
            plt.plot(x_vals2, fx, linewidth=0.3, label='')

            plt.ticklabel_format(axis='y', style='sci', scilimits=(0, 4))
            plt.legend(loc=3, bbox_to_anchor=(0, 0.8))
            plt.title('$%s$ vs $T$' % self.display_symbol)
            plt.xlabel('$T$ (K)')

            plt.ylabel('$%s$ (%s)' % (self.display_symbol, self.units))

            fig = plt.gcf()
            pdf.savefig(fig)
            plt.close()

        if show:
            webbrowser.open_new(path)


if __name__ == "__main__":
    import unittest
    from polynomial_test import *
    unittest.main()



#!/usr/bin/env python3
"""
This module contains helper functions to ease various tasks.
"""


import os
from datetime import datetime


def get_path_relative_to_module(module_file_path, relative_target_path):
    """
    Calculate a path relative to the specified module file.

    :param module_file_path: The file path to the module.
    """
    module_path = os.path.dirname(module_file_path)
    path = os.path.join(module_path, relative_target_path)
    path = os.path.abspath(path)
    return path


def get_date(date):
    """
    Get the date from a value that could be a date object or a string.

    :param date: The date object or string.

    :returns: The date object.
    """
    if type(date) is str:
        return datetime.strptime(date, '%Y-%m-%d').date()
    else:
        return date


if __name__ == "__main__":
    import unittest
    from helpers_test import HelpersUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
This module contains all the code used to test the testee module.
"""


import unittest
from os import remove
from os.path import realpath, dirname, join, isfile

from auxi.tools.materialphysicalproperties import polynomial as testee
from auxi.tools.materialphysicalproperties.core import DataSet


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


MODULE_PATH = dirname(realpath(__file__))


class PolynomialModelTTester(unittest.TestCase):
    """
    The tester for the PolynomialModelT class.
    """

    def _test_properties(self, model):
        self.assertEqual(model.material, 'Air')
        self.assertEqual(model.property, 'Density')
        self.assertEqual(model.symbol, 'rho')
        self.assertEqual(model.display_symbol, '\\rho')
        self.assertEqual(model.references, None)
        self.assertEqual(model.datasets, ['dataset-air-lienhard2015.csv'])

        coeffs = [2.207010112413834e-28, -1.8498386015487013e-24,
                  6.761238643515948e-21, -1.415594789179995e-17,
                  1.875696238757754e-14, -1.6406192076125332e-11,
                  9.591604317302061e-09, -3.714488392719117e-06,
                  0.0009239972475093591, -0.13882710545123356,
                  11.147387712425617]

        for c, cref in zip(model._coeffs, coeffs):
            self.assertAlmostEqual(c, cref)

    def test_create(self):
        """
        Test whether a model is created successfully from a data set.
        """
        dataset_path = join(MODULE_PATH, 'data/dataset-air-lienhard2015.csv')
        dataset = DataSet(dataset_path)
        model = testee.PolynomialModelT.create(dataset, 'rho', 10)
        self._test_properties(model)

    def test_read(self):
        """
        Test whether a model is created successfully by loading it from a json
        file.
        """
        model = testee.PolynomialModelT.read(join(MODULE_PATH,
                                             'data/air-rho.json'))
        self._test_properties(model)

    def test_write(self):
        """
        Test whether a model is successfully written to a json file.
        """
        dataset_path = join(MODULE_PATH, 'data/dataset-air-lienhard2015.csv')
        dataset = DataSet(dataset_path)

        json_path = join(MODULE_PATH, 'test.json')

        model = testee.PolynomialModelT.create(dataset, 'rho', 10)
        model.write(json_path)
        self.assertTrue(isfile(json_path))
        model = testee.PolynomialModelT.read(json_path)
        remove(json_path)
        self._test_properties(model)

    def test_construct(self):
        """
        Test whether a model is constructed successfully.
        """

        file_path = join(MODULE_PATH, 'data/air-rho.json')

        model = testee.PolynomialModelT.read(file_path)
        testee.PolynomialModelT(model.material, model.property, model.symbol,
                                model.display_symbol, model.units,
                                model.references, model.datasets,
                                model._coeffs)
        self._test_properties(model)

    def test_calculate(self):
        """
        Test whether the property value is calculated successfully.
        """
        file_path = join(MODULE_PATH, 'data/air-rho.json')
        model = testee.PolynomialModelT.read(file_path)

        T = 100.0
        # self.assertEqual(model.calculate(T=T), 3.6026669128620208)
        self.assertEqual(model.calculate(T=T), 3.6049798036305774)
        T = 200.0
        self.assertEqual(model.calculate(T=T), 1.7581685372604081)
        T = 300.0
        self.assertEqual(model.calculate(T=T), 1.1791874628593089)
        T = 400.0
        self.assertEqual(model.calculate(T=T), 0.87918802181468259)
        T = 500.0
        self.assertEqual(model.calculate(T=T), 0.70171039180794637)
        T = 600.0
        self.assertEqual(model.calculate(T=T), 0.59118824960721206)

    def test_plot(self):
        """
        Test whether the plots are created.
        """
        dataset_path = join(MODULE_PATH, 'data/dataset-air-lienhard2015.csv')
        dataset = DataSet(dataset_path)
        model = testee.PolynomialModelT.create(dataset, 'rho', 10)

        pdf_path = join(MODULE_PATH, 'test.pdf')

        model.plot(dataset, pdf_path)
        self.assertTrue(isfile(pdf_path))
        remove(pdf_path)

    def test_call(self):
        """
        Test whether the property value is calculated successfully via the
        __call__ magic method.
        """
        file_path = join(MODULE_PATH, 'data/air-rho.json')
        model = testee.PolynomialModelT.read(file_path)

        T = 100.0
        self.assertEqual(model(T=T), 3.6026669128620208)
        T = 200.0
        self.assertEqual(model(T=T), 1.7581685372604081)
        T = 300.0
        self.assertEqual(model(T=T), 1.1791874628593089)
        T = 400.0
        self.assertEqual(model(T=T), 0.87918802181468259)
        T = 500.0
        self.assertEqual(model(T=T), 0.70171039180794637)
        T = 600.0
        self.assertEqual(model(T=T), 0.59118824960721206)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides testing code for the psdslurrymaterial module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
import os
import numpy
from auxi.modelling.process.materials import slurry
from auxi.modelling.process.materials.slurry import Material, MaterialPackage

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


# =============================================================================
# Types.
# =============================================================================

class SlurryMaterialUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.process.materials.slurry.Material class.
    """

    def setUp(self):
        self.material = Material(
            "material",
            os.path.join(slurry.DEFAULT_DATA_PATH,
                         r"psdslurrymaterial.test.materiala.txt"))

    def test_constructor(self):
        self.assertEqual(self.material.name, "material")
        self.assertEqual(len(self.material.size_classes), 10)
        self.assertEqual(self.material.size_class_count, 10)
        self.assertEqual(len(self.material.assays), 5)

    def test_get_size_class_index(self):
        self.assertEqual(self.material.get_size_class_index(307.2E-3), 0)
        self.assertEqual(self.material.get_size_class_index(38.4E-3), 2)
        self.assertEqual(self.material.get_size_class_index(600.0E-6), 6)
        self.assertEqual(self.material.get_size_class_index(0.0E0), 9)

    def test_create_empty_assay(self):
        empty_assay = self.material.create_empty_assay()
        self.assertEqual(len(empty_assay), 10)
        self.assertEqual(empty_assay.sum(), 0.0)

    def test_add_assay(self):
        new_assay = self.material.create_empty_assay()
        new_assay[0] = 0.5
        new_assay[2] = 0.5
        self.material.add_assay("new_assay", 0.3, 0.05, new_assay)
        self.assertTrue(
            numpy.all(self.material.assays["new_assay"] == new_assay))
        self.assertEqual(self.material.solid_densities["new_assay"], 0.3)
        self.assertEqual(self.material.H2O_fractions["new_assay"], 0.05)

    def test_get_assay_total(self):
        self.assertEqual(self.material.get_assay_total("DryFeedA"),
                         1.0000000000000002)
        self.assertEqual(self.material.get_assay_total("DryMillCharge"),
                         0.99)

    def test_create_package(self):
        package = self.material.create_package("DryFeedA", 123.456, True)
        self.assertAlmostEqual(package.get_mass(), 123.45599999999999,
                               places=10)


class SlurryMaterialPackageUnitTester(unittest.TestCase):

    def setUp(self):
        self.materiala = Material(
            "materiala",
            os.path.join(slurry.DEFAULT_DATA_PATH,
                         r"psdslurrymaterial.test.materiala.txt"))
        self.materiala_package_a = self.materiala.create_package("DryFeedA",
                                                                 1234.5, True)
        self.materiala_package_b = self.materiala.create_package(
            "DryMillCharge",
            2345.6, True)

    def test_constructor(self):
        size_class_masses = self.materiala.assays["DryFeedA"] * 123.4 / \
                          self.materiala.assays["DryFeedA"].sum()
        package = MaterialPackage(self.materiala, 0.3, 0.05, size_class_masses)
        self.assertEqual(package.material, self.materiala)
        self.assertAlmostEqual(package.get_mass(), 123.45, places=10)
        self.assertEqual(package.solid_density, 0.3)
        self.assertEqual(package.H2O_mass, 0.05)
        self.assertTrue(numpy.all(package.size_class_masses ==
                                  size_class_masses))

    def test___str__(self):
        self.assertGreater(len(self.materiala_package_a.__str__()), 0)

    def test_add_operator_1(self):
        """
        other = MaterialPackage
        Test whether the add operator calculates the resulting package
        correctly.
        """

        package_a_plus_b = self.materiala_package_a + self.materiala_package_b

        self.assertAlmostEqual(self.materiala_package_a.get_mass(), 1234.5)

        self.assertAlmostEqual(self.materiala_package_b.get_mass(),
                               2345.5999999999995)

        self.assertAlmostEqual(package_a_plus_b.get_mass(), 3580.0999999999999)

    # def test_add_operator_2(self):
        # mix_package = self.mix.create_package(None, 0.0)
        # mix_package = mix_package + self.materiala_package_a
        # mix_package = mix_package + self.reductant_package_a

        # self.assertEqual(mix_package.get_mass(),
        #                  self.materiala_package_a.get_mass() +
        #                  self.reductant_package_a.get_mass())

        # self.assertRaises(Exception, self.add_incompatible_packages)

    # def add_incompatible_packages(self):
        # result = self.materiala_package_a + self.reductant_package_a
        # result = result * 1.0

    def test_add_operator_3(self):
        """
        other = tuple (size_class, mass)
        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        packageAplus75micron = self.materiala_package_a + (75.0E-6, 123.4)

        self.assertAlmostEqual(packageAplus75micron.get_mass(),
                               1357.8999999999999)

    def test_extract_1(self):
        temp_package_a = self.materiala_package_a.clone()
        mass = 432.1
        diff_package = temp_package_a.extract(mass)

        self.assertAlmostEqual(temp_package_a.get_mass(),
                               self.materiala_package_a.get_mass() - mass)

        self.assertAlmostEqual(diff_package.get_mass(), mass)

    def test_subtract_operator_2(self):
        temp_package_a = self.materiala_package_a.clone()
        size_class = 4.8E-3
        mass = 123.4
        diff_package = temp_package_a.extract((size_class, mass))

        self.assertAlmostEqual(
            temp_package_a.get_mass(),
            self.materiala_package_a.get_mass() - mass)
        self.assertAlmostEqual(
            temp_package_a.get_size_class_mass(size_class),
            self.materiala_package_a.get_size_class_mass(size_class) - mass)

        self.assertAlmostEqual(diff_package.get_mass(), mass)
        self.assertAlmostEqual(
            diff_package.get_size_class_mass(size_class),
            mass)

    def test_subtract_operator_3(self):
        temp_package_a = self.materiala_package_a.clone()
        size_class = 4.8E-3
        mass = temp_package_a.get_size_class_mass(size_class)
        diff_package = temp_package_a.extract(str(size_class))

        self.assertAlmostEqual(
            temp_package_a.get_mass(),
            self.materiala_package_a.get_mass() - mass)
        self.assertAlmostEqual(
            temp_package_a.get_size_class_mass(size_class),
            self.materiala_package_a.get_size_class_mass(size_class) -
            mass)

        self.assertAlmostEqual(diff_package.get_mass(), mass)
        self.assertAlmostEqual(
            diff_package.get_size_class_mass(size_class),
            mass)

    def test_multiply_operator(self):
        temp_package_a = self.materiala_package_a.clone()

        mul_package_1 = temp_package_a * 0.0
        self.assertEqual(mul_package_1.get_mass(), 0.0)
        self.assertTrue(numpy.all(mul_package_1.size_class_masses ==
                                  temp_package_a.size_class_masses * 0.0))

        mul_package_2 = temp_package_a * 1.0
        self.assertEqual(mul_package_2.get_mass(), temp_package_a.get_mass())
        self.assertTrue(numpy.all(mul_package_2.size_class_masses ==
                                  temp_package_a.size_class_masses))

        mul_package_2 = temp_package_a * 123.4
        self.assertAlmostEqual(
            mul_package_2.get_mass(),
            temp_package_a.get_mass() * 123.4)
        self.assertTrue(numpy.all(mul_package_2.size_class_masses ==
                                  temp_package_a.size_class_masses * 123.4))

    def test_clone(self):
        clone = self.materiala_package_a.clone()

        self.assertEqual(clone.get_mass(), self.materiala_package_a.get_mass())
        self.assertTrue(numpy.all(clone.size_class_masses ==
                                  self.materiala_package_a.size_class_masses))

    def test_get_mass(self):
        self.assertAlmostEqual(self.materiala_package_a.get_mass(), 1234.5)
        self.assertAlmostEqual(
            self.materiala_package_b.get_mass(),
            2345.5999999999995)

    def test_get_assay(self):
        self.assertTrue(
            numpy.all(self.materiala_package_a.get_assay() -
                      self.materiala.assays["DryFeedA"] /
                      self.materiala.assays["DryFeedA"].sum() < 1.0E-16))
        self.assertTrue(
            numpy.all(self.materiala_package_a.get_assay() -
                      self.materiala.assays["DryFeedA"] /
                      self.materiala.assays["DryFeedA"].sum() > -1.0E-16))

    def test_get_size_class_mass(self):
        assay = "DryFeedA"
        size_class = 4.8E-3
        for size_class in self.materiala.size_classes:
            index = self.materiala.get_size_class_index(size_class)
            mass = 1234.5 * self.materiala.assays[assay][index] / \
                self.materiala.get_assay_total(assay)
            self.assertEqual(
                self.materiala_package_a.get_size_class_mass(size_class),
                mass)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module contains all the code used to test the testee module.
"""


import unittest
import csv
from os import remove
from os.path import realpath, dirname, join, isfile

from auxi.tools.materialphysicalproperties import core as testee


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


MODULE_PATH = dirname(realpath(__file__))


class DataSetTester(unittest.TestCase):
    """
    The tester for the DataSet class.
    """

    def test_create_template(self):
        mat_name = 'test'
        template_path = MODULE_PATH
        testee.DataSet.create_template(mat_name, template_path)
        file_path = join(MODULE_PATH, 'dataset-%s.csv' % mat_name)
        self.assertTrue(isfile(file_path))

        with open(file_path, newline='') as file:
            lines = csv.reader(file, delimiter=',', quotechar='"')
            ix = 0
            for l in lines:
                if ix == 0:
                    self.assertEqual(l[0], 'Name')
                    self.assertEqual(l[1], 'test')
                if ix == 7:
                    self.assertEqual(l[0], '100.0')
                    self.assertEqual(l[1], '0.0')
                    self.assertEqual(l[2], '10.0')
                    self.assertEqual(l[3], '100.0')
                if ix == 16:
                    self.assertEqual(l[0], '550.0')
                    self.assertEqual(l[1], '9.0')
                    self.assertEqual(l[2], '19.0')
                    self.assertEqual(l[3], '109.0')
                ix += 1
        remove(file_path)

    def _create_dataset(self):
        path = join(dirname(realpath(__file__)),
                    'data/dataset-air-lienhard2015.csv')
        return testee.DataSet(path)

    def test_construct(self):
        """
        Test whether a data set is constructed successfully from a file.
        """
        self._create_dataset()

    def test_name(self):
        """
        Test whether the data set name is successfully loaded.
        """
        ds = self._create_dataset()
        self.assertEqual(ds.name, 'dataset-air-lienhard2015.csv')

    def test_material(self):
        """
        Test whether the material name is successfully loaded.
        """
        ds = self._create_dataset()
        self.assertEqual(ds.material, 'Air')

    def test_description(self):
        """
        Test whether the data set description is successfully loaded.
        """
        ds = self._create_dataset()
        self.assertEqual(ds.description, 'Thermophysical properties of air at'
                         ' 1 atm pressure.')

    def test_reference(self):
        """
        Test whether the data set reference is successfully loaded.
        """
        ds = self._create_dataset()
        self.assertEqual(ds.reference, 'lienhard2015')

    def test_parameter_names(self):
        """
        Test whether the data set parameter names are successfully loaded.
        """
        ds = self._create_dataset()
        self.assertEqual(ds.col_names,
                         ['Temperature', 'Density', 'Heat Capacity',
                          'Dynamic Viscosity', 'Kinematic Viscosity',
                          'Thermal Conductivity', 'Thermal Diffusivity',
                          'Prandtl Number'])

    def test_parameter_units(self):
        """
        Test whether the data set parameter units are successfully loaded.
        """
        ds = self._create_dataset()
        self.assertEqual(ds.col_units,
                         ['K', 'kg/m3', 'J/kg/K', 'kg/m/s', 'm2/s', 'W/m/K',
                          'm2/s', 'None'])

    def test_parameter_symbols(self):
        """
        Test whether the data set parameter symbols are successfully loaded.
        """
        ds = self._create_dataset()
        self.assertEqual(ds.col_symbols,
                         ['T', 'rho', 'Cp', 'mu', 'nu', 'k', 'alpha', 'Pr'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides physical property data sets and models for gases.
"""

from sys import modules
from os.path import realpath, dirname, join

from auxi.tools.materialphysicalproperties.core import DataSet
from auxi.tools.materialphysicalproperties.polynomial import PolynomialModelT
from auxi.tools.materialphysicalproperties.idealgas import \
    BetaT as IgBetaT, RhoT as IgRhoT
from auxi.modelling.process.materials.core import Material
from auxi.tools.materialphysicalproperties.core import StateOfMatter


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman, Marno Grewar'
__credits__ = ['Christoff Kok', 'Johan Zietsman', 'Marno Grewar']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


def _path(relative_path):
    path = modules[__name__].__file__
    path = realpath(path)
    path = dirname(path)
    return join(path, relative_path)


def _create_polynomial_model(symbol, degree):
    newmod = PolynomialModelT.create(air_dataset, symbol, degree)
    newmod.plot(air_dataset, _path('temp.pdf'), False)
    newmod.write(_path('data/air-%s.json' % symbol.lower()))


air_dataset = DataSet(_path('data/dataset-air-lienhard2015.csv'))

# _create_polynomial_model('Cp', 14)
# _create_polynomial_model('k', 8)
# _create_polynomial_model('mu', 8)
# _create_polynomial_model('rho', 14)

air_dict = {}
air_dict['rho'] = IgRhoT(28.9645, 101325.0)
air_dict['Cp'] = PolynomialModelT.read(_path(r'data/air-cp.json'))
air_dict['mu'] = PolynomialModelT.read(_path(r'data/air-mu.json'))
air_dict['k'] = PolynomialModelT.read(_path(r'data/air-k.json'))
air_dict['beta'] = IgBetaT()

air = Material("Air", StateOfMatter.gas, air_dict)



#!/usr/bin/env python3
"""
This module provides functions to calculate dimensionless quantities used when
doing transport phenomena calculations.
"""


# TODO: Add mathematical formula to each function. See Gr for example.
# TODO: Test to make sure that math formulae appear correctly on readthedocs.
# TODO: Test to make sure that math formulae appear correctly in LaTeX.
# DECISION: Use the most compact set of parameters for each function, as long
#   as
#          it does not cause ambiguity.
#          For example,
#              'nu' instead of 'mu' and 'rho'
#              'Ts' and 'Tinf' instead of 'dT'


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman, Marno Grewar'
__credits__ = ['Christoff Kok', 'Johan Zietsman', 'Marno Grewar']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


def Gr(L: float, Ts: float, Tf: float, beta: float, nu: float, g: float):
    """
    Calculate the Grashof number.

    :param L: [m] heat transfer surface characteristic length.
    :param Ts: [K] heat transfer surface temperature.
    :param Tf: [K] bulk fluid temperature.
    :param beta: [1/K] fluid coefficient of thermal expansion.
    :param nu: [m2/s] fluid kinematic viscosity.

    :returns: float

    .. math::
        \\mathrm{Gr} = \\frac{g \\beta (Ts - Tinf ) L^3}{\\nu ^2}

    Characteristic dimensions:
        * vertical plate: vertical length
        * pipe: diameter
        * bluff body: diameter
    """

    return g * beta * (Ts - Tf) * L**3.0 / nu**2.0


def Pr(nu: float, alpha: float) -> float:
    """
    Calculate the Prandtl number.

    :param nu: [m2/s] fluid kinematic viscosity / momentum diffusivity.
    :param alpha: [m2/s] fluid thermal diffusivity.

    :returns: float
    """

    return nu / alpha


def Re(L: float, v: float, nu: float) -> float:
    """
    Calculate the Reynolds number.

    :param L: [m] surface characteristic length.
    :param v: [m/s] fluid velocity relative to the object.
    :param nu: [m2/s] fluid kinematic viscosity.

    :returns: float
    """

    return v * L / nu


def Ra(L: float, Ts: float, Tf: float, alpha: float, beta: float, nu: float
       ) -> float:
    """
    Calculate the Grashof number.

    :param L: [m] heat transfer surface characteristic length.
    :param Ts: [K] heat transfer surface temperature.
    :param Tf: [K] bulk fluid temperature.
    :param alpha: [m2/s] fluid thermal diffusivity.
    :param beta: [1/K] fluid coefficient of thermal expansion.
    :param nu: [m2/s] fluid kinematic viscosity.

    :returns: float

    Ra = Gr*Pr

    Characteristic dimensions:
        * vertical plate: vertical length
        * pipe: diameter
        * bluff body: diameter
    """

    return g * beta * (Ts - Tinf) * L**3.0 / (nu * alpha)


def Nu(L: float, h: float, k: float) -> float:
    """
    Calculate the Nusselt number.

    :param L: [m] heat transfer surface characteristic length.
    :param h: [W/K/m2] convective heat transfer coefficient.
    :param k: [W/K/m] fluid thermal conductivity.

    :returns: float
    """

    return h * L / k


def Sh(L: float, h: float, D: float) -> float:
    """
    Calculate the Sherwood number.

    :param L: [m] mass transfer surface characteristic length.
    :param h: [m/s] mass transfer coefficient.
    :param D: [m2/s] fluid mass diffusivity.

    :returns: float
    """

    return h * L / D



#!/usr/bin/env python3
"""
This module provides material and material package classes that can do
size distribution and slurry calculations.
"""

import os
import sys
import copy

import numpy

from auxi.core.objects import Object
from auxi.core.objects import NamedObject

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


# TODO: Johan, I replaced 'self.psds' with 'self.assays'
#   to match the psd.py module.


class Material(NamedObject):
    """
    Represents a particulate material consisting of multiple particle size
    classes.

    :param name: The material's name.
    :param file_path: The path of the material definition file.
    :param description: the material's description

    The format of the text file is as follows:

    * The lines are space separated. The values in a line are separated by \
      one or more spaces.
    * The first line is a heading line.
    * The second line contains the density of the solid material.
    * The third line contains the water fraction of the slurry (wet basis).
    * All subsequent lines contain a particle size, followed by mass \
      fractions (dry basis).
    * Particle sizes are indicated in [meter].
    * The first column lists the particle sizes in the material. Each class \
      must be interpreted as "mass fraction retained". In other words if the \
      size class is indicated as 307.2E-3, it means that it is the class of \
      material retained on a 307.2mm screen, and can also be though of as \
      +307.2mm material. The first size class represents the largest \
      particles. The final size class should be zero, as it represents \
      all material that passed through the smallest aperture screen.
    * All subsequent columns describe assays of the material.

    The following is an example of a material text file::

        SizeClass       DryFeedA  DryMillCharge  WetFeedA  WetMillCharge  Water
        solid_density   3.00      3.00           3.00      3.00           1.0.
        H2O             0.00      0.00           0.80      0.60           1.00
        307.2E-3        0.20      0.02           0.20      0.02           0.00
        108.6E-3        0.18      0.06           0.18      0.06           0.00
        38.4E-3         0.17      0.04           0.17      0.04           0.00
        13.6E-3         0.07      0.03           0.07      0.03           0.00
        4.8E-3          0.13      0.03           0.13      0.03           0.00
        1.7E-3          0.07      0.04           0.07      0.04           0.00
        600.0E-6        0.06      0.18           0.06      0.18           0.00
        210.0E-6        0.02      0.50           0.02      0.50           0.00
        75.0E-6         0.10      0.09           0.10      0.09           0.00
        0.0E0           0.00      0.00           0.00      0.00           0.00
    """

    def __init__(self, name, file_path, description=None):
        # Initialise the material's properties.
        self.name = name
        self.size_classes = list()

        # Read the material's data from the file and prepare it for use.
        f = open(file_path, "r")
        lines = f.readlines()
        f.close()
        lines = self._prepare_lines(lines)

        # Determine the assay names, and create a dictionary entry
        # for each assay.
        assay_names = lines[0].split(" ")
        del(assay_names[0:1])
        self.assays = dict()
        self.solid_densities = dict()
        self.H2O_fractions = dict()
        for assay_name in assay_names:
            self.assays[assay_name] = numpy.array([])
            self.solid_densities[assay_name] = 1.0
            self.H2O_fractions[assay_name] = 0.0

        # Read the solid densities of the assays.
        strings = lines[1].split(" ")
        if not strings[0] == "solid_density":
            raise Exception(
                "Invalid data file. "
                "The second line of the data file must start with"
                "'solid_density'.")
        # Add the solid densities.
        for j in range(0, len(self.solid_densities)):
            assay_name = assay_names[j]
            self.solid_densities[assay_name] = float(strings[j+1])

        # Read the water fractions of the assays.
        strings = lines[2].split(" ")
        if not strings[0] == "H2O":
            raise Exception(
                "Invalid data file. "
                "The third line of the data file must start with 'H2O'.")
        # Add the water fractions.
        for j in range(0, len(self.H2O_fractions)):
            assay_name = assay_names[j]
            self.H2O_fractions[assay_name] = float(strings[j+1])

        # Read the size classes and mass fractions.
        for i in range(3, len(lines)):
            strings = lines[i].split(" ")
            # Not a full line.
            if len(strings) < len(assay_names) + 1:
                continue
            # Add the new size class.
            self.size_classes.append(float(strings[0]))
            # Add the mass fractions to the assays.
            for j in range(0, len(self.assays)):
                assay_name = assay_names[j]
                self.assays[assay_name] = numpy.append(
                    self.assays[assay_name], float(strings[j+1]))

        # Initialise the remaining properties.
        self.size_class_count = len(self.size_classes)

    def __str__(self):
        """
        Create a string representation of self.
        """

        result = "Material: name='" + self.name + "'\n"

        # Create the header line of the table.
        result = result + "Compound".ljust(20)
        assay_names = sorted(self.assays.keys())
        for assay_name in assay_names:
            result = result + assay_name.ljust(20)
        result = result + "\n"

        # Create the solid density line.
        result = result + "Solid density".ljust(20)
        assay_names = sorted(self.assays.keys())
        for assay_name in assay_names:
            result = result + str(self.solid_densities[assay_name]).ljust(20)
        result = result + "\n"

        # Create the H2O fraction line.
        result = result + "H2O fraction".ljust(20)
        assay_names = sorted(self.assays.keys())
        for assay_name in assay_names:
            result = result + str(self.H2O_fractions[assay_name]).ljust(20)
        result = result + "\n"

        # Create the content lines of the table.
        for size_class in self.size_classes:
            result = result + str(size_class).ljust(20)
            compound_index = self.get_size_class_index(size_class)
            for assay_name in assay_names:
                result = result + str(
                    self.assays[assay_name][compound_index]).ljust(20)
            result = result + "\n"
        return result

    def _prepare_lines(self, lines):
        """
        Prepare the lines read from the text file before starting to process
        it.
        """

        result = list()
        for line in lines:
            # Remove all whitespace characters (e.g. spaces, line breaks, etc.)
            # from the start and end of the line.
            line = line.strip()
            # Replace all tabs with spaces.
            line = line.replace("\t", " ")
            # Replace all repeating spaces with a single space.
            while line.find("  ") > -1:
                line = line.replace("  ", " ")
            result.append(line)
        return result

    def get_size_class_index(self, size_class):
        """
        Determine the index of the specified size class.

        :param size_class: The formula and phase of the specified size class,
          e.g. 'Fe2O3[S1]'.

        :returns: The index of the specified size class.
        """

        return self.size_classes.index(size_class)

    def create_empty_assay(self):
        """
        Create an empty array to store an assay. The array's length will be
        equal to the number of size classes in the material.

        :returns: A floating point array.
        """

        return numpy.zeros(self.size_class_count)

    '''
    # TODO: Johan I replaced this method with the method above.
    def create_empty_psd(self):
        """
        Create an empty array to store a psd. The array's length will be equal
        to the number of size classes in the material.

        :returns: A floating point array.
        """

        return numpy.zeros(self.size_class_count)
    '''

    def add_assay(self, name, solid_density, H2O_fraction, assay):
        """Add an assay to the material.

        :param name: The name of the new assay.
        :param assay: A numpy array containing the size class mass fractions
          for the assay. The sequence of the assay's elements must correspond
          to the sequence of the material's size classes.
        """

        if not type(solid_density) is float:
            raise Exception("Invalid solid density. It must be a float.")
        self.solid_densities[name] = solid_density

        if not type(H2O_fraction) is float:
            raise Exception("Invalid H2O fraction. It must be a float.")
        self.H2O_fractions[name] = H2O_fraction

        if not type(assay) is numpy.ndarray:
            raise Exception("Invalid assay. It must be a numpy array.")
        elif not assay.shape == (self.size_class_count,):
            raise Exception(
                "Invalid assay: It must have the same number of elements as "
                "the material has size classes.")
        elif name in self.assays.keys():
            raise Exception(
                "Invalid assay: An assay with that name already exists.")
        self.assays[name] = assay

    def get_assay_total(self, name):
        """
        Calculate the total of the specified assay.

        :param name: The name of the assay.

        :returns: The total mass fraction of the specified assay.
        """

        return sum(self.assays[name])

    '''
    # TODO: Johan I replaced this method with the method above.
    def get_psd_total(self, name):
        """Calculate the total of the specified assay's psd.
        :param name: The name of the assay.

        :returns: The total mass fraction of the specified assay.
        """

        return sum(self.assays[name])
    '''

    def create_package(self, assay=None, mass=0.0, normalise=True):
        """
        Create a MaterialPackage based on the specified parameters.

        :param assay: The name of the assay based on which the package
          must be created.
        :param mass: [kg] The mass of the package.
        :param normalise: Indicates whether the assay must be normalised
          before creating the package.

        :returns: The created MaterialPackage.
        """

        if assay is None:
            return MaterialPackage(self, 1.0, 0.0, self.create_empty_assay())

        if normalise:
            assay_total = self.get_assay_total(assay)
            if assay_total == 0.0:
                assay_total = 1.0
        else:
            assay_total = 1.0
        H2O_mass = mass * self.H2O_fractions[assay]
        solid_mass = mass - H2O_mass
        return MaterialPackage(self,
                               self.solid_densities[assay],
                               H2O_mass,
                               solid_mass * self.assays[assay] / assay_total)


class MaterialPackage(Object):
    """
    A package of a slurry material consisting of multiple particle size
    classes.

    :param material: A reference to the Material to which self belongs.
    :param size_class_masses: [kg] [kg] The masses of the size classes in the
          package.
    """

    def __init__(self, material, solid_density, H2O_mass, size_class_masses):
        # Confirm that the parameters are OK.
        if not type(material) is Material:
            raise TypeError(
                "Invalid material type. Must be psdslurrymaterial.Material")
        if not type(size_class_masses) is numpy.ndarray:
            raise TypeError(
                "Invalid size_class_masses type. Must be numpy.ndarray.")

        # Initialise the object's properties.
        self.material = material
        self.solid_density = solid_density
        self.H2O_mass = H2O_mass
        self.size_class_masses = size_class_masses

    def __str__(self):
        """
        Create a string representation of the object.
        """

        result = "MaterialPackage\n"
        result = result + "material".ljust(24) + self.material.name + "\n"
        result = result + "mass fraction solids".ljust(24) + \
            str(self.get_mass_fraction_solids()) + "\n"
        result = result + "volume fraction solids".ljust(24) + \
            str(self.get_volume_fraction_solids()) + "\n"
        result = result + "solid density".ljust(24) + \
            str(self.solid_density) + "\n"
        result = result + "slurry density".ljust(24) + \
            str(self.get_density()) + "\n"
        result = result + "mass".ljust(24) + str(self.get_mass()) + "\n"
        result = result + "H2O mass".ljust(24) + str(self.H2O_mass) + "\n"
        result = result + "volume".ljust(24) + str(self.get_volume()) + "\n"
        result = result + "Component masses:\n"
        for size_class in self.material.size_classes:
            index = self.material.get_size_class_index(size_class)
            result = result + str(size_class).ljust(24) + str(
                self.size_class_masses[index]) + "\n"
        return result

    def __add__(self, other):
        """
        Addition operator (+).
        Add self and 'other' together, return the result as a new package,
        and leave self unchanged.

        :param other: Can can be one of the following:
          1. MaterialPackage: 'other' is added to self to create a new package.
          2. tuple: (size class, mass) The specified mass of the specified size
          class is added to self.

        :returns: A new Material package that is the sum of self and 'other'.
        """

        # Add another package.
        if type(other) is MaterialPackage:
            solid_mass = self.get_solid_mass()
            other_solid_mass = other.get_solid_mass()
            solid_density = (solid_mass + other_solid_mass) / \
                (solid_mass / self.solid_density +
                    other_solid_mass / other.solid_density)
            H2O_mass = self.H2O_mass + other.H2O_mass
            # Packages of the same material.
            if self.material == other.material:
                result = MaterialPackage(
                    self.material,
                    solid_density,
                    H2O_mass,
                    self.size_class_masses + other.size_class_masses)
                return result
            else:  # Packages of different materials.
                result = self.clone()
                result.solid_density = solid_density
                result.H2O_mass = H2O_mass
                for size_class in other.material.size_classes:
                    if size_class not in self.material.size_classes:
                        raise Exception(
                            "Packages of '" + other.material.name +
                            "' cannot be added to packages of '" +
                            self.material.name + "'. The size class '" +
                            size_class + "' was not found in '" +
                            self.material.name + "'.")
                    result = result + (
                        size_class, other.get_size_class_mass(size_class))
                return result

        # Add the specified mass of water.
        elif self._is_H2O_mass_tuple(other):
            # Added material variables.
            mass = other[1]

            # Create the result package.
            result = self.clone()
            result.H2O_mass += mass
            return result

        # Add the specified mass of the specified size class.
        elif self._is_size_class_mass_tuple(other):
            # Added material variables.
            size_class = other[0]
            compound_index = self.material.get_size_class_index(size_class)
            mass = other[1]

            # Create the result package.
            result = self.clone()
            result.size_class_masses[compound_index] = \
                result.size_class_masses[compound_index] + mass
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid addition argument.")

    def __mul__(self, scalar):
        """
        The multiplication operator (*).
        Create a new package by multiplying self with other.

        :param scalar: The result is a new package with its content equal to
          self multiplied by a scalar, leaving self unchanged.

        :returns: A new MaterialPackage equal to self package multiplied by
          other.
        """

        # Multiply with a scalar floating point number.
        if type(scalar) is float or \
                type(scalar) is numpy.float64 or \
                type(scalar) is numpy.float32:
            if scalar < 0.0:
                raise Exception(
                    "Invalid multiplication operation. "
                    "Cannot multiply package with negative number.")
            result = MaterialPackage(
                self.material,
                self.solid_density,
                self.H2O_mass * scalar, self.size_class_masses * scalar)
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid multiplication argument.")

    def _is_size_class_mass_tuple(self, value):
        """
        Determines whether value is a tuple of the format
        (size class(float), mass(float)).

        :param value: The value to check.

        :returns: Whether the value is a tuple in the required format.
        """

        if not type(value) is tuple:
            return False
        elif not len(value) == 2:
            return False
        elif not type(value[0]) is float:
            return False
        elif not type(value[1]) is float and \
                not type(value[1]) is numpy.float64 and \
                not type(value[1]) is numpy.float32:
            return False
        else:
            return True

    def _is_H2O_mass_tuple(self, value):
        """
        Determines whether value is a tuple of the format
        (size class(float), mass(float)).

        :param value: The value to check.

        :returns: Whether the value is a tuple in the required format.
        """

        if not type(value) is tuple:
            return False
        elif not len(value) == 2:
            return False
        elif not type(value[0]) is str and not value[0] == "H2O":
            return False
        elif not type(value[1]) is float and \
                not type(value[1]) is numpy.float64 and \
                not type(value[1]) is numpy.float32:
            return False
        else:
            return True

    def clone(self):
        """
        Create a complete copy of self.

        :returns: A MaterialPackage that is identical to self.
        """

        result = copy.copy(self)
        result.size_class_masses = copy.deepcopy(self.size_class_masses)
        return result

    def clear(self):
        """
        Set all the size class masses and H20_mass in the package to zero
        and the solid_density to 1.0
        """

        self.solid_density = 1.0
        self.H2O_mass = 0.0
        self.size_class_masses = self.size_class_masses * 0.0

    def get_assay(self):
        """
        Determine the assay of self.

        :returns: [mass fractions] An array containing the assay of self.
        """

        return self.size_class_masses / self.size_class_masses.sum()

    def get_mass(self):
        """
        Determine the mass of self.

        :returns: [kg] The mass of self.
        """

        return self.size_class_masses.sum() + self.H2O_mass

    def get_solid_mass(self):
        """
        Determine the solid mass of self.

        :returns: [kg] The solid mass of self.
        """

        return self.size_class_masses.sum()

    def get_size_class_mass(self, size_class):
        """
        Determine the mass of the specified size class in self.

        :param size_class: The formula and phase of the size class,
          e.g. 'Fe2O3[S1]'

        :returns: [kg] The mass of the size class in self.
        """

        return self.size_class_masses[self.material.get_size_class_index(
            size_class)]

    # TODO: Test
    def get_size_class_mass_fraction(self, size_class):
        """
        Determine the mass fraction of the specified size class in self.

        :param size_class: The formula and phase of the size class,
          e.g. Fe2O3[S1]

        :returns: The mass fraction of the size class in self.
        """

        return self.get_size_class_mass(size_class) / self.get_solid_mass()

    def get_density(self):
        """
        Determine the density of self.
        """

        return self.get_mass() / self.get_volume()

    def get_mass_fraction_solids(self):
        """
        Determine the mass fraction of the solids of self.
        """

        return self.get_solid_mass() / self.get_mass()

    def get_volume(self):
        """
        Determine the volume of self.
        """

        return self.H2O_mass / 1.0 + self.get_solid_mass() / self.solid_density

    def get_volume_fraction_solids(self):
        """
        Determine the volume fraction of the solids of self.
        """

        return 1.0 - (self.H2O_mass / 1.0) / self.get_volume()

    def extract(self, other):
        """
        Extract 'other' from self, modifying self and returning the extracted
        material as a new package.

        :param other: Can be one of the following:

          * float: A mass equal to other is extracted from self. Self is
            reduced by other and the extracted package is returned as a new
            package.
          * tuple (size class, mass): The other tuple specifies the mass of a
            size class to be extracted. It is extracted from self and the
            extracted mass is returned as a new package.
          * string: The 'other' string specifies the size class to be
            extracted. All of the mass of that size class will be removed from
            self and a new package created with it.


        :returns: A new material package containing the material that was
          extracted from self.
        """

        # Extract the specified mass.
        if type(other) is float or \
                type(other) is numpy.float64 or \
                type(other) is numpy.float32:
            if other > self.get_mass():
                raise Exception(
                    "Invalid extraction operation. "
                    "Cannot extract a mass larger than the package's mass.")
            fraction_to_subtract = other / self.get_mass()
            result = MaterialPackage(
                self.material,
                self.solid_density,
                self.H2O_mass * fraction_to_subtract,
                self.size_class_masses * fraction_to_subtract)
            self.H2O_mass = self.H2O_mass * (1.0 - fraction_to_subtract)
            self.size_class_masses = \
                self.size_class_masses * (1.0 - fraction_to_subtract)
            return result

        # Extract the specified mass of water.
        elif self._is_H2O_mass_tuple(other):
            if other[1] > self.H2O_mass:
                raise Exception(
                    "Invalid extraction operation. "
                    "Cannot extract a water mass larger than what the package "
                    "contains.")
            self.H2O_mass = self.H2O_mass - other[1]
            resultarray = self.size_class_masses * 0.0
            result = MaterialPackage(
                self.material, self.solid_density, other[1], resultarray)
            return result

        # Extract the specified mass of the specified size class.
        elif self._is_size_class_mass_tuple(other):
            index = self.material.get_size_class_index(other[0])
            if other[1] > self.size_class_masses[index]:
                raise Exception(
                    "Invalid extraction operation. "
                    "Cannot extract a size class mass larger than what the "
                    "package contains.")
            self.size_class_masses[index] = \
                self.size_class_masses[index] - other[1]
            resultarray = self.size_class_masses*0.0
            resultarray[index] = other[1]
            result = MaterialPackage(
                self.material, self.solid_density, 0.0, resultarray)
            return result

        # Extract all of the specified size class.
        elif type(other) is str:
            index = self.material.get_size_class_index(float(other))
            result = self * 0.0
            result.size_class_masses[index] = self.size_class_masses[index]
            self.size_class_masses[index] = 0.0
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid extraction argument.")

    # TODO: Test
    # TODO: Document
#    def add_to(self, other):
#        # Add another package.
#        if type(other) is MaterialPackage:
#            # Packages of the same material.
#            if self.material == other.material:
#                self.size_class_masses = \
#                   self.size_class_masses + other.size_class_masses
#            else: # Packages of different materials.
#                for size_class in other.material.size_classes:
#                    if size_class not in self.material.size_classes:
#                        raise Exception(
#                           "Packages of '" + other.material.name +
#                           "' cannot be added to packages of '" +
#                           self.material.name + "'. The size class '" +
#                           size_class + "' was not found in '" +
#                           self.material.name + "'.")
#                    self.add_to(
#                        (size_class, other.get_size_class_mass(size_class)))
#
#        # Add the specified mass of the specified size class.
#        elif self._is_size_class_mass_tuple(other):
#            # Added material variables.
#            size_class = other[0]
#            compound_index = self.material.get_size_class_index(size_class)
#            mass = other[1]
#
#            # Create the result package.
#            self.size_class_masses[compound_index] = \
#                self.size_class_masses[compound_index] + mass
#
#        # If not one of the above, it must be an invalid argument.
#        else:
#            raise TypeError("Invalid addition argument.")


def _get_default_data_path():
    module_path = os.path.dirname(sys.modules[__name__].__file__)
    data_path = os.path.join(module_path, r"data")
    data_path = os.path.abspath(data_path)
    return data_path

DEFAULT_DATA_PATH = _get_default_data_path()



#!/usr/bin/env python3
"""
This module provides physical property data sets and models for slags.
"""

from sys import modules
from os.path import realpath, dirname, join
from math import exp

from auxi.tools.materialphysicalproperties.core import Model, DataSet
from auxi.tools.chemistry.stoichiometry import amount_fractions

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Markus Erwee, Johan Zietsman'
__credits__ = ['Markus Erwee', 'Johan Zietsman']
__maintainer__ = 'Johan zietsman'
__email__ = 'johan.zietsman@ex-mente.co.za'
__status__ = 'Planning'


def _path(relative_path):
    path = modules[__name__].__file__
    path = realpath(path)
    path = dirname(path)
    return join(path, relative_path)


class UrbainViscosityTx(Model):
    """
    A model that describes the variation in the dynamic viscosity of liquid
    slag as a function of temperature and composition expressed in mole
    fraction.
    """

    def __init__(self):
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0},
                        'x': {'required': True, 'type': 'dict'}}
        super().__init__('Slag', 'Dynamic Viscosity', 'mu', '\\mu', 'Pa.s',
                         state_schema, None, None)

    def calculate(self, **state):
        """
        Calculate dynamic viscosity at the specified temperature and
        composition:

        :param T: [K] temperature
        :param x: [mole fraction] composition dictionary , e.g. \
        {'SiO2': 0.25, 'CaO': 0.25, 'MgO': 0.25, 'FeO': 0.25}

        :returns: [Pa.s] dynamic viscosity

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """

        T = state['T']
        x = state['x']

        # normalise mole fractions
        x_total = sum(x.values())
        x = {compound: x[compound]/x_total for compound in x.keys()}

        xg = x.get('SiO2', .00) + x.get('P2O5', 0.0)
        xm = x.get('CaO', 0.0) + x.get('MgO', 0.0) + x.get('Na2O', 0.0) + \
            x.get('K2O', 0.0) + 3.0*x.get('CaF2', 0.0) + x.get('FeO', 0.0) + \
            x.get('MnO', 0.0) + 2.0*x.get('TiO2', 0.0) + 2.0*x.get('ZrO2', 0.0)

        xa = x.get('Al2O3', 0.0) + x.get('Fe2O3', 0.0) + x.get('B2O3', 0.0)

        # Note 2*XFeO1.5 = XFe2O3

        norm = 1.0 + x.get('CaF2', 0.0) + x.get('Fe2O3', 0.0) + \
            x.get('TiO2', 0.0) + x.get('ZrO2', 0.0)

        xg_norm = xg / norm
        xm_norm = xm / norm
        xa_norm = xa / norm

        alpha = xm_norm / (xm_norm + xa_norm)

        B0 = 13.8 + 39.9355*alpha - 44.049*alpha**2.0
        B1 = 30.481 - 117.1505*alpha + 129.9978*alpha**2.0
        B2 = -40.9429 + 234.0846*alpha - 300.04*alpha**2.0
        B3 = 60.7619 - 153.9276*alpha + 211.1616*alpha**2.0

        B = B0 + B1*xg_norm + B2*xg_norm**2.0 + B3*xg_norm**3.0

        A = exp(-0.2693*B - 11.6725)

        result = A*T*exp(1000.0*B/T)  # [P]

        return result / 10.0  # [Pa.s]


class UrbainViscosityTy(UrbainViscosityTx):
    """
    A model that describes the variation in the dynamic viscosity of liquid
    slag as a function of temperature and composition expressed in mass
    fraction.
    """

    def __init__(self):
        super().__init__()
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0},
                        'y': {'required': True, 'type': 'dict'}}
        self._create_validator(state_schema)

    def calculate(self, **state):
        """
        Calculate dynamic viscosity at the specified temperature and
        composition:

        :param T: [K] temperature
        :param y: [mass fraction] composition dictionary , e.g. \
        {'SiO2': 0.25, 'CaO': 0.25, 'MgO': 0.25, 'FeO': 0.25}

        :returns: [Pa.s] dynamic viscosity

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """

        T = state['T']
        y = state['y']
        x = amount_fractions(y)
        return super().calculate(T=T, x=x)


class RiboudViscosityTx(Model):
    """
    A model that describes the variation in the dynamic viscosity of liquid
    slag as a function of temperature and composition expressed in mole
    fraction.
    """

    def __init__(self):
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0},
                        'x': {'required': True, 'type': 'dict'}}
        super().__init__('Slag', 'Dynamic Viscosity', 'mu', '\\mu', 'Pa.s',
                         state_schema, None, None)

    def calculate(self, **state):
        """
        Calculate dynamic viscosity at the specified temperature and
        composition:

        :param T: [K] temperature
        :param x: [mole fraction] composition dictionary , e.g. \
        {'SiO2': 0.25, 'CaO': 0.25, 'MgO': 0.25, 'FeO': 0.25}

        :returns: [Pa.s] dynamic viscosity

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """

        T = state['T']
        x = state['x']

        # create the slag constituent categories
        compounds_sio2 = ['SiO2', 'PO2.5', 'TiO2', 'ZrO2']
        compounds_cao = ['CaO', 'MgO', 'FeO1.5', 'FeO', 'MnO', 'BO1.5']
        compounds_al2o3 = ['Al2O3']
        compounds_caf2 = ['CaF2']
        compounds_na2o = ['Na2O', 'K2O']
        compounds_all = (compounds_sio2 + compounds_cao + compounds_al2o3 +
                         compounds_caf2 + compounds_na2o)

        # convert compounds with two cations to single cation equivalents
        if 'P2O5' in x:
            x['PO2.5'] = 2.0 * x['P2O5']
        if 'Fe2O3' in x:
            x['FeO1.5'] = 2.0 * x['Fe2O3']
        if 'B2O3' in x:
            x['BO1.5'] = 2.0 * x['B2O3']

        # normalise mole fractions, use only compounds in compounds_all
        x_total = sum([x.get(c, 0.0) for c in compounds_all])
        x = {c: x.get(c, 0.0)/x_total for c in compounds_all}

        # calculate the cateogry mole fractions
        x1 = sum([x.get(c, 0.0) for c in compounds_sio2])
        x2 = sum([x.get(c, 0.0) for c in compounds_cao])
        x3 = sum([x.get(c, 0.0) for c in compounds_al2o3])
        x4 = sum([x.get(c, 0.0) for c in compounds_caf2])
        x5 = sum([x.get(c, 0.0) for c in compounds_na2o])

        # TODO: Why is x1 not used? This looks suspicious.
        A = exp(-17.51 + 1.73*x2 + 5.82*x4 + 7.02*x5 - 33.76*x3)
        B = 31140.0 - 23896.0*x2 - 46356.0*x4 - 39159.0*x5 + 68833.0*x3

        result = A*T*exp(B/T)  # [P]

        return result / 10.0  # [Pa.s]


class RiboudViscosityTy(RiboudViscosityTx):
    """
    A model that describes the variation in the dynamic viscosity of liquid
    slag as a function of temperature and composition expressed in mass
    fraction.
    """

    def __init__(self):
        super().__init__()
        state_schema = {'T': {'required': True, 'type': 'float', 'min': 0.0},
                        'y': {'required': True, 'type': 'dict'}}
        self._create_validator(state_schema)

    def calculate(self, **state):
        """
        Calculate dynamic viscosity at the specified temperature and
        composition:

        :param T: [K] temperature
        :param y: [mass fraction] composition dictionary , e.g. \
        {'SiO2': 0.25, 'CaO': 0.25, 'MgO': 0.25, 'FeO': 0.25}

        :returns: [Pa.s] dynamic viscosity

        The **state parameter contains the keyword argument(s) specified above\
        that are used to describe the state of the material.
        """

        T = state['T']
        y = state['y']
        x = amount_fractions(y)
        return super().calculate(T=T, x=x)


ds1 = DataSet(_path('data/dataset-slag-slagatlas1995.csv'))



#!/usr/bin/env python3
"""
This module provides classes to manage time.
"""

from datetime import datetime, timedelta

from enum import Enum
from dateutil.relativedelta import relativedelta

from auxi.core.objects import NamedObject

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class TimePeriod(Enum):
    """
    Represents a period in time.
    """

    millisecond = 1
    second = 2
    minute = 3
    hour = 4
    day = 5
    week = 6
    month = 7
    year = 8


class Clock(NamedObject):
    """
    Represents a clock that provices functions to manage a ticking clock based
    on a time period as well as retreive the current tick's date since the
    start date.
    """

    def __init__(self, name, description=None,
                 start_datetime=datetime.min,
                 timestep_period_duration=TimePeriod.month,
                 timestep_period_count=1):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param start_datetime: The start datetime.
        :param timestep_period_duration: The duration of each time period.
        :param timestep_period_count: The number of periods that makes up a
          timestep.
        """

        super(Clock, self).__init__(name, description)
        self.start_datetime = start_datetime
        self.timestep_period_duration = timestep_period_duration
        self.timestep_period_count = timestep_period_count
        self.timestep_ix = 0

    def tick(self):
        """
        Increment the clock's timestep index.
        """

        self.timestep_ix += 1

    def reset(self):
        """
        Resets the clock's timestep index to '0'.
        """

        self.timestep_ix = 0

    def get_datetime_at_period_ix(self, ix):
        """
        Get the datetime at a given period.

        :param period: The index of the period.

        :returns: The datetime.
        """

        if self.timestep_period_duration == TimePeriod.millisecond:
            return self.start_datetime + timedelta(milliseconds=ix)
        elif self.timestep_period_duration == TimePeriod.second:
            return self.start_datetime + timedelta(seconds=ix)
        elif self.timestep_period_duration == TimePeriod.minute:
            return self.start_datetime + timedelta(minutes=ix)
        elif self.timestep_period_duration == TimePeriod.hour:
            return self.start_datetime + timedelta(hours=ix)
        elif self.timestep_period_duration == TimePeriod.day:
            return self.start_datetime + relativedelta(days=ix)
        elif self.timestep_period_duration == TimePeriod.week:
            return self.start_datetime + relativedelta(days=ix*7)
        elif self.timestep_period_duration == TimePeriod.month:
            return self.start_datetime + relativedelta(months=ix)
        elif self.timestep_period_duration == TimePeriod.year:
            return self.start_datetime + relativedelta(years=ix)

    def get_datetime(self):
        """
        Get the clock's current datetime.

        :returns: The datetime.
        """

        return self.get_datetime_at_period_ix(self.timestep_ix)

if __name__ == "__main__":
    import unittest
    from auxi.core.time.time_test import ClockUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
This module contains all the code used to test the testee module.
"""


import unittest
import random

from auxi.tools.materialphysicalproperties import idealgas as testee
from auxi.tools.physicalconstants import R
from auxi.tools.chemistry.stoichiometry import molar_mass as mm


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class BetaTTester(unittest.TestCase):
    """
    The tester for the BetaT class.
    """

    def test_construct(self):
        """
        Test whether a model is constructed successfully.
        """
        testee.BetaT()

    def test_calculate(self):
        """
        Test whether the property value is calculated successfully.
        """
        model = testee.BetaT()
        T = random.uniform(0.0, 3000.0)
        self.assertEqual(model.calculate(T=T), 1.0/T)
        T = random.uniform(0.0, 3000.0)
        self.assertEqual(model.calculate(T=T), 1.0/T)
        T = random.uniform(0.0, 3000.0)
        self.assertEqual(model.calculate(T=T), 1.0/T)

    def test_call(self):
        """
        Test whether the property value is calculated successfully via the
        __call__ magic method.
        """
        model = testee.BetaT()
        T = random.uniform(0.0, 3000.0)
        self.assertEqual(model(T=T), 1.0/T)
        T = random.uniform(0.0, 3000.0)
        self.assertEqual(model(T=T), 1.0/T)
        T = random.uniform(0.0, 3000.0)
        self.assertEqual(model(T=T), 1.0/T)


class RhoTTester(unittest.TestCase):
    """
    The tester for the RhoT class.
    """

    def test_create(self):
        """
        Test whether a model is created successfully.
        """
        mm = random.uniform(1.0, 200.0)
        P = random.uniform(1.0, 101325.0*5.0)

        testee.RhoT(mm, P)

    def test_calculate(self):
        """
        Test whether the property value is calculated successfully.
        """
        mm = random.uniform(1.0, 200.0)
        P = random.uniform(1.0, 101325.0*5.0)

        model = testee.RhoT(mm, P)

        T = random.uniform(0.0, 3000.0)
        self.assertAlmostEqual(model.calculate(T=T), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        self.assertAlmostEqual(model.calculate(T=T), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        self.assertAlmostEqual(model.calculate(T=T), mm*P/R/T/1000.0)

    def test_call(self):
        """
        Test whether the property value is calculated successfully via the
        __call__ magic method.
        """
        mm = random.uniform(1.0, 200.0)
        P = random.uniform(1.0, 101325.0*5.0)

        model = testee.RhoT(mm, P)

        T = random.uniform(0.0, 3000.0)
        self.assertAlmostEqual(model(T=T), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        self.assertAlmostEqual(model(T=T), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        self.assertAlmostEqual(model(T=T), mm*P/R/T/1000.0)


class RhoTPTester(unittest.TestCase):
    """
    The tester for the RhoTP class.
    """

    def test_create(self):
        """
        Test whether a model is created successfully.
        """
        mm = random.uniform(1.0, 200.0)

        testee.RhoTP(mm)

    def test_calculate(self):
        """
        Test whether the property value is calculated successfully.
        """
        mm = random.uniform(1.0, 200.0)

        model = testee.RhoTP(mm)

        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        self.assertAlmostEqual(model.calculate(T=T, P=P), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        self.assertAlmostEqual(model.calculate(T=T, P=P), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        self.assertAlmostEqual(model.calculate(T=T, P=P), mm*P/R/T/1000.0)

    def test_call(self):
        """
        Test whether the property value is calculated successfully via the
        __call__ magic method.
        """
        mm = random.uniform(1.0, 200.0)

        model = testee.RhoTP(mm)

        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        self.assertAlmostEqual(model(T=T, P=P), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        self.assertAlmostEqual(model(T=T, P=P), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        self.assertAlmostEqual(model(T=T, P=P), mm*P/R/T/1000.0)


class RhoTPxTester(unittest.TestCase):
    """
    The tester for the RhoTPx class.
    """

    def _create_x_and_mm(self):
        x = {}

        x['N2'] = random.uniform(0.0, 10.0)
        x['O2'] = random.uniform(0.0, 10.0)
        x['Ar'] = random.uniform(0.0, 10.0)
        x['H2O'] = random.uniform(0.0, 10.0)

        total = sum(x.values())

        x = {key: x[key]/total for key in x.keys()}

        molar_mass = sum([x[key]*mm(key) for key in x.keys()])

        return x, molar_mass

    def test_create(self):
        """
        Test whether a model is created successfully.
        """
        testee.RhoTPx()

    def test_calculate(self):
        """
        Test whether the property value is calculated successfully.
        """
        model = testee.RhoTPx()

        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        x, mm = self._create_x_and_mm()
        self.assertAlmostEqual(model.calculate(T=T, P=P, x=x), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        x, mm = self._create_x_and_mm()
        self.assertAlmostEqual(model.calculate(T=T, P=P, x=x), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        x, mm = self._create_x_and_mm()
        self.assertAlmostEqual(model.calculate(T=T, P=P, x=x), mm*P/R/T/1000.0)

    def test_call(self):
        """
        Test whether the property value is calculated successfully via the
        __call__ magic method.
        """
        model = testee.RhoTPx()

        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        x, mm = self._create_x_and_mm()
        self.assertAlmostEqual(model(T=T, P=P, x=x), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        x, mm = self._create_x_and_mm()
        self.assertAlmostEqual(model(T=T, P=P, x=x), mm*P/R/T/1000.0)
        T = random.uniform(0.0, 3000.0)
        P = random.uniform(1.0, 101325.0*5.0)
        x, mm = self._create_x_and_mm()
        self.assertAlmostEqual(model(T=T, P=P, x=x), mm*P/R/T/1000.0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides classes representing the accounting double entry system.
"""

from datetime import datetime

from enum import Enum

from auxi.core.objects import NamedObject
from auxi.core.reporting import ReportFormat


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class AccountType(Enum):
    """
    Represents the type of general ledger account.
    """

    asset = 1
    equity = 2
    expense = 3
    liability = 4
    revenue = 5

AT = AccountType


class GeneralLedgerAccount(NamedObject):
    """
    Represents an account of a general ledger.

    :param name: The name.
    :param description: The description.
    :param number: The number.
    :param account_type: The type of account.
    """

    def __init__(self, name, description=None,
                 number=None,
                 account_type=AccountType.revenue):
        self._parent_path = ""
        self.path = name
        self.number = number
        self.account_type = account_type
        self.accounts = []
        super(GeneralLedgerAccount, self).__init__(name, description)

    def __getitem__(self, key):
        return [a for a in self.accounts if a.name == key][0]

    def __missing__(self, key):
        return None

    def set_parent_path(self, value):
        """
        Set the parent path and the path from the new parent path.

        :param value: The path to the object's parent
        """

        self._parent_path = value
        self.path = value + r'/' + self.name
        self._update_childrens_parent_path()

    def _update_childrens_parent_path(self):
        for a in self.accounts:
            a.set_parent_path(self.path)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        if self._parent_path != "":
            self.path = self._parent_path + r'/' + self.name
        else:
            self.path = self.name
        self._update_childrens_parent_path()

    def create_account(self, name, number=None, description=None):
        """
        Create a sub account in the account.

        :param name: The account name.
        :param description: The account description.
        :param number: The account number.

        :returns: The created account.
        """

        new_account = GeneralLedgerAccount(name, description, number,
                                           self.account_type)
        new_account.set_parent_path(self.path)
        self.accounts.append(new_account)
        return new_account

    def remove_account(self, name):
        """
        Remove an account from the account's sub accounts.

        :param name: The name of the account to remove.
        """

        acc_to_remove = None
        for a in self.accounts:
            if a.name == name:
                acc_to_remove = a
        if acc_to_remove is not None:
            self.accounts.remove(acc_to_remove)

    def get_child_account(self, account_name):
        """
        Retrieves a child account.
        This could be a descendant nested at any level.

        :param account_name: The name of the account to retrieve.

        :returns: The child account, if found, else None.
        """

        if(r'/' in account_name):
            accs_in_path = account_name.split(r'/', 1)

            curr_acc = self[accs_in_path[0]]
            if curr_acc is None:
                return None
            return curr_acc.get_child_account(accs_in_path[1])
            pass
        else:
            return self[account_name]
        return None


class Transaction(NamedObject):
    """
    Represents a financial transaction between two general ledger accounts.

    :param name: The name.
    :param description: The description.
    :param tx_date: The transaction's date.
    :param dt_account: The account to debit.
    :param cr_account: The account to credit.
    :param source: The source that created the transaction.
    :param amount: The transaction's amount.
    :param is_closing_dt_account: Specifies wether this is a closing debit
      account.
    :param is_closing_cr_account: Specifies wether this is a closing credit
      account.
    """

    def __init__(self, name, description=None,
                 tx_date=datetime.min.date(),
                 dt_account=None, cr_account=None,
                 source=None, amount=0.00,
                 is_closing_dt_account=False,
                 is_closing_cr_account=False):
        super(Transaction, self).__init__(name, description)
        self.tx_date = tx_date
        self.dt_account = dt_account
        self.cr_account = cr_account
        self.source = source
        self.amount = amount
        self.is_closing_dt_account = is_closing_dt_account
        self.is_closing_cr_account = is_closing_cr_account


class TransactionTemplate(NamedObject):
    """
    Represents a template for how a transaction is to be created.

    :param name: The name of the transaction.
    :param description: The description of the transaction.
    :param dt_account: The account to debit.
    :param cr_account: The account to credit.
    """

    def __init__(self, name, dt_account, cr_account, description=None):
        """
        """
        super(TransactionTemplate, self).__init__(name, description)
        self.dt_account = dt_account
        self.cr_account = cr_account


class GeneralLedgerStructure(NamedObject):
    """
    The account structure of a general ledger.

    :param name: The name.
    :param description: The description.
    """

    # TODO: add _acc_ name abbr. descriptions

    def __init__(self, name, description=None):
        super(GeneralLedgerStructure, self).__init__(name, description)
        self.accounts = []

        ca = self._create_account_

        # create income statement account categories
        self._acci_unalloc_inc_ = ca("Unallocated Income Statement", "I05",
                                     AT.revenue)
        self._acci_sales_ = ca("Sales", "I10", AT.revenue)
        self._acci_cos_ = ca("Cost of Sales", "I15", AT.expense)
        self._acci_oth_inc_ = ca("Other Income", "I20", AT.revenue)
        self._acci_exp_ = ca("Expense", "I25", AT.expense)
        self._acci_tax_ = ca("Tax", "I30", AT.expense)
        self._acci_divs_ = ca("Dividends", "I35", AT.expense)

        # create balance sheet account categories
        self._accb_sh_cap_ = ca("Share Capital", "B10", AT.asset)
        self._accb_ret_inc_ = ca("Retained Income", "B15", AT.equity)
        self._accb_shh_loan_ = ca("Shareholders Loan", "B20", AT.liability)
        self._accb_lt_borrow_ = ca("Long Term Borrowing", "B25", AT.liability)
        self._accb_oth_lt_lia_ = ca("Other Long Term Liabilities", "B30",
                                    AT.liability)
        self._accb_fixed_ass_ = ca("Fixed Assets", "B35", AT.asset)
        self._accb_investments_ = ca("Investments", "B40", AT.asset)
        self._accb_oth_fixed_ass_ = ca("Other Fixed Assets", "B45", AT.asset)
        self._accb_inventory_ = ca("Inventory", "B50", AT.asset)
        self._accb_acc_receivable_ = ca("Accounts Receivable", "B55", AT.asset)
        self._accb_bank_ = ca("Bank", "B60", AT.asset)
        self._accb_bank_.create_account("Default", "0000")
        self._accb_oth_curr_ass_ = ca("Other Current Assets", "B65", AT.asset)
        self._accb_acc_payable_ = ca("Account Payable", "B70", AT.liability)
        self._accb_taxation_ = ca("Taxation", "B75", AT.liability)
        self._accb_oth_curr_lia_ = ca("Other Current Liabilities", "B80",
                                      AT.liability)

        # create the year end sub accounts
        self._acci_gross_prof_ = self._acci_unalloc_inc_.create_account(
            "Gross Profit", "0000")
        self._acci_inc_sum_ = self._acci_unalloc_inc_.create_account(
            "Income Summary", "0010")
        self._accb_ret_earnings_acc_ = self._accb_ret_inc_.create_account(
            "Retained Earnings", "0000")
        self._acci_inc_tax_exp_acc_ = self._acci_tax_.create_account(
            "Income Tax Expense", "0000")

        self.tax_payment_account = "Bank/Default"

    def __getitem__(self, key):
        items = [a for a in self.accounts if a.name == key]
        if len(items) == 0:
            return None
        else:
            return items[0]

    def __missing__(self, key):
        return None

    def _create_account_(self, name, number, account_type):
        """
        Create an account in the general ledger structure.

        :param name: The account name.
        :param number: The account number.
        :param account_type: The account type.

        :returns: The created account.
        """

        new_acc = GeneralLedgerAccount(name, None, number, account_type)
        self.accounts.append(new_acc)
        return new_acc

    def get_account(self, account_name):
        """
        Retrieves an account from the general ledger structure
        given the account name.

        :param account_name: The account name.

        :returns: The requested account, if found, else None.
        """

        if(r'/' in account_name):
            accs_in_path = account_name.split(r'/', 1)

            curr_acc = self[accs_in_path[0]]
            if curr_acc is None:
                return None
            return curr_acc.get_child_account(accs_in_path[1])
            pass
        else:
            return self[account_name]
        return None

    def get_account_decendants(self, account):
        """
        Retrieves an account's decendants from the general ledger structure
        given the account name.

        :param account_name: The account name.

        :returns: The decendants of the account.
        """

        result = []
        for child in account.accounts:
            self._get_account_and_decendants_(child, result)
        return result

    def _get_account_and_decendants_(self, account, result):
        """
        Returns the account and all of it's sub accounts.

        :param account: The account.
        :param result: The list to add all the accounts to.
        """

        result.append(account)
        for child in account.accounts:
            self._get_account_and_decendants_(child, result)

    def validate_account_names(self, names):
        """
        Validates wether the accounts in a list of account names exists.

        :param names: The names of the accounts.

        :returns: The decendants of the account.
        """

        for name in names:
            if self.get_account(name) is None:
                raise ValueError("The account '{}' does not exist in the"
                                 "  general ledger structure.".format(name))

    def report(self, format=ReportFormat.printout, output_path=None):
        """
        Returns a report of this class.

        :param format: The format of the report.
        :param output_path: The path to the file the report is written to.
          If None, then the report is not written to a file.

        :returns: The decendants of the account.
        """

        rpt = GlsRpt(self, output_path)
        return rpt.render(format)


class GeneralLedger(NamedObject):
    """
    Represents the account structure of a general ledger.

    :param name: The name.
    :param structure: The general ledger structure.
    :param description: The description.
    """

    def __init__(self, name, structure, description=None):
        super(GeneralLedger, self).__init__(name, description)
        self.structure = structure
        self.transactions = []

    def create_transaction(self, name, description=None,
                           tx_date=datetime.min.date(),
                           dt_account=None, cr_account=None,
                           source=None, amount=0.00):
        """
        Create a transaction in the general ledger.

        :param name: The transaction's name.
        :param description: The transaction's description.
        :param tx_date: The date of the transaction.
        :param cr_account: The transaction's credit account's name.
        :param dt_account: The transaction's debit account's name.
        :param source: The name of source the transaction originated from.
        :param amount: The transaction amount.

        :returns: The created transaction.
        """

        new_tx = Transaction(name, description, tx_date,
                             dt_account, cr_account, source, amount)
        self.transactions.append(new_tx)
        return new_tx

    def transaction_list(self, start=datetime.min,
                         end=datetime.max,
                         format=ReportFormat.printout,
                         component_path="",
                         output_path=None):
        """
        Generate a transaction list report.

        :param start: The start date to generate the report for.
        :param end: The end date to generate the report for.
        :param format: The format of the report.
        :param component_path: The path of the component to filter the report's
          transactions by.
        :param output_path: The path to the file the report is written to.
          If None, then the report is not written to a file.

        :returns: The generated report.
        """

        rpt = TransactionList(self, start, end, component_path, output_path)
        return rpt.render(format)

    def balance_sheet(self, end=datetime.max,
                      format=ReportFormat.printout, output_path=None):
        """
        Generate a transaction list report.

        :param end: The end date to generate the report for.
        :param format: The format of the report.
        :param output_path: The path to the file the report is written to.
          If None, then the report is not written to a file.

        :returns: The generated report.
        """

        rpt = BalanceSheet(self, end, output_path)
        return rpt.render(format)

    def income_statement(self, start=datetime.min,
                         end=datetime.max,
                         format=ReportFormat.printout,
                         component_path="",
                         output_path=None):
        """
        Generate a transaction list report.

        :param start: The start date to generate the report for.
        :param end: The end date to generate the report for.
        :param format: The format of the report.
        :param component_path: The path of the component to filter the report's
          transactions by.
        :param output_path: The path to the file the report is written to.
          If None, then the report is not written to a file.

        :returns: The generated report.
        """

        rpt = IncomeStatement(self, start, end, component_path, output_path)
        return rpt.render(format)


from auxi.modelling.financial.reporting import GeneralLedgerStructure as GlsRpt
from auxi.modelling.financial.reporting import TransactionList
from auxi.modelling.financial.reporting import BalanceSheet
from auxi.modelling.financial.reporting import IncomeStatement


from auxi.modelling.financial.reporting import GeneralLedgerStructure as GlsRpt
from auxi.modelling.financial.reporting import TransactionList as TxList


if __name__ == "__main__":
    import unittest
    from auxi.modelling.financial.des_test import GeneralLedgerAccountUnitTester
    from auxi.modelling.financial.des_test import TransactionUnitTester
    from auxi.modelling.financial.des_test import TransactionTemplateUnitTester
    from auxi.modelling.financial.des_test import GeneralLedgerStructureUnitTester
    from auxi.modelling.financial.des_test import GeneralLedgerUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides an classes used to create a business structure.
"""

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from collections import OrderedDict

from auxi.core.objects import NamedObject
from auxi.modelling.financial.des import AccountType
from auxi.modelling.financial.des import GeneralLedger


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Activity(NamedObject):
    """
    Represents an activity base class. An activity will typically
    represent a transaction activity in a business.

    :param name: The name.
    :param description: The description.
    :param start: The datetime the activity should be started.
    :param end: The datetime the activity should be run until.
    :param interval: The interval of the activity.
    """

    def __init__(self, name,
                 start=datetime.min, end=datetime.max,
                 interval=1,
                 description=None):
        self._parent_path = ""
        super(Activity, self).__init__(name, description)
        self.start_datetime = self._get_date_(start)
        self.end_datetime = self._get_date_(end)
        self.interval = interval
        self.start_period_ix = -1
        self.end_period_ix = -1
        self.period_count = -1

    def _get_date_(self, date):
        if type(date) is str:
            return datetime.strptime(date, '%Y-%m-%d')
        else:
            return date

    def set_parent_path(self, value):
        """
        Set the parent path and the path from the new parent path.

        :param value: The path to the object's parent
        """

        self._parent_path = value
        self.path = value + r'/' + self.name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.path = self._parent_path + r'/' + self.name

    def _meet_execution_criteria(self, ix_period):
        if self.interval != 0 and (ix_period+1) % self.interval != 0:
            return False
        return ix_period >= self.start_period_ix and \
            ix_period + self.interval  <= self.end_period_ix

    def prepare_to_run(self, clock, period_count):
        """
        Prepare the activity for execution.

        :param clock: The clock containing the execution start time and
          execution period information.
        :param period_count: The total amount of periods this activity will be
          requested to be run for.
        """

        if self.start_period_ix == -1 and self.start_datetime != datetime.min:
            # Set the Start period index
            for i in range(0, period_count):
                if clock.get_datetime_at_period_ix(i) > self.start_datetime:
                    self.start_period_ix = i
                    break
        if self.start_period_ix == -1:
            self.start_period_ix = 0
        if self.period_count == -1 and self.end_datetime != datetime.max:
            # Set the Start date
            for i in range(0, period_count):
                if clock.get_datetime_at_period_ix(i) > self.end_datetime:
                    self.period_count = i - self.start_period_ix
                    break
        if self.period_count != -1:
            self.end_period_ix = self.start_period_ix + self.period_count
        else:
            self.end_period_ix = self.start_period_ix + period_count

    def get_referenced_accouts(self):
        """
        Retrieve the general ledger accounts referenced in this instance.

        :returns: The referenced accounts.
        """

        return []


class Component(NamedObject):
    """
    Represents an component class. A component class that represents a
    component of an entity. A component has business activities

    :param name: The name.
    :param description: The description.
    """

    def __init__(self, name, gl, description=None):
        self._parent_path = ""
        self.path = name
        self.gl = gl
        self.components = []
        self.activities = []
        super(Component, self).__init__(name, description=description)

    def __getitem__(self, name):
        return self.get_component(name)

    def __call__(self, name):
        return self.get_activity(name)

    def set_parent_path(self, value):
        """
        Set the parent path and the path from the new parent path.

        :param value: The path to the object's parent.
        """

        self._parent_path = value
        self.path = value + r'/' + self.name
        self._update_childrens_parent_path()

    def _update_childrens_parent_path(self):
        for c in self.components:
            c.set_parent_path(self.path)
        for a in self.activities:
            a.set_parent_path(self.path)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.path = self._parent_path + r'/' + self.name
        self._update_childrens_parent_path()

    def create_component(self, name, description=None):
        """
        Create a sub component in the business component.

        :param name: The new component's name.
        :param description: The new component's description.

        :returns: The created component.
        """

        new_comp = Component(name, self.gl, description=description)
        new_comp.set_parent_path(self.path)
        self.components.append(new_comp)
        return new_comp

    def remove_component(self, name):
        """
        Remove a sub component from the component.

        :param name: The name of the component to remove.
        """

        component_to_remove = None
        for c in self.components:
            if c.name == name:
                component_to_remove = c
        if component_to_remove is not None:
            self.components.remove(component_to_remove)

    def get_component(self, name):
        """
        Retrieve a child component given its name.

        :param name: The name of the component.

        :returns: The component.
        """

        return [c for c in self.components if c.name == name][0]

    def add_activity(self, activity):
        """
        Add an activity to the component.

        :param activity: The activity.
        """

        self.gl.structure.validate_account_names(
            activity.get_referenced_accounts())
        self.activities.append(activity)
        activity.set_parent_path(self.path)

    def get_activity(self, name):
        """
        Retrieve an activity given its name.

        :param name: The name of the activity.

        :returns: The activity.
        """

        return [a for a in self.activities if a.name == name][0]

    def prepare_to_run(self, clock, period_count):
        """
        Prepare the component for execution.

        :param clock: The clock containing the execution start time and
          execution period information.
        :param period_count: The total amount of periods this activity will be
          requested to be run for.
        """

        for c in self.components:
            c.prepare_to_run(clock, period_count)
        for a in self.activities:
            a.prepare_to_run(clock, period_count)

    def run(self, clock, generalLedger):
        """
        Execute the component at the current clock cycle.

        :param clock: The clock containing the current execution time and
          period information.
        :param generalLedger: The general ledger into which to create the
          transactions.
        """

        for c in self.components:
            c.run(clock, generalLedger)
        for a in self.activities:
            a.run(clock, generalLedger)


class Entity(NamedObject):
    """
    Represents an entity class. An entity consists of business components
    e.g. Sales department. It executes its components and performs
    financial year end transcations.

    :param name: The name.
    :param gl_structure: The general ledger structure the entity's
      general ledger will be initialized with.
    :param description: The description.
    :param period_count: The number of periods the entity should be run for.
    """

    def __init__(self, name, gl_structure, description=None):
        self.gl = GeneralLedger(
            "General Ledger",
            gl_structure,
            description="General Ledger")
        self._parent_path = ""
        self.path = name
        self.components = []
        self.negative_income_tax_total = 0
        self._prev_year_end_datetime = datetime.min
        self._curr_year_end_datetime = datetime.min
        self._exec_year_end_datetime = datetime.min
        self.period_count = -1
        super(Entity, self).__init__(name, description=description)

    def __getitem__(self, key):
        return [c for c in self.components if c.name == key][0]

    def set_parent_path(self, value):
        """
        Set the parent path and the path from the new parent path.

        :param value: The path to the object's parent
        """

        self._parent_path = value
        self.path = value + r'/' + self.name
        self._update_childrens_parent_path()

    def _update_childrens_parent_path(self):
        for c in self.components:
            c.set_parent_path(self.path)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.path = self._parent_path + r'/' + self.name
        self._update_childrens_parent_path()

    def create_component(self, name, description=None):
        """
        Create a component in the business entity.

        :param name: The component's name.
        :param description: The component's description.

        :returns: The created component.
        """

        new_comp = Component(name, self.gl, description=description)
        new_comp.set_parent_path(self.path)
        self.components.append(new_comp)
        return new_comp

    def remove_component(self, name):
        """
        Remove a component from the entity.

        :param name: The name of the component to remove.
        """

        component_to_remove = None
        for c in self.components:
            if c.name == name:
                component_to_remove = c
        if component_to_remove is not None:
            self.components.remove(component_to_remove)

    def prepare_to_run(self, clock, period_count):
        """
        Prepare the entity for execution.

        :param clock: The clock containing the execution start time and
          execution period information.
        :param period_count: The total amount of periods this activity will be
          requested to be run for.
        """

        self.period_count = period_count

        self._exec_year_end_datetime = clock.get_datetime_at_period_ix(
            period_count)
        self._prev_year_end_datetime = clock.start_datetime
        self._curr_year_end_datetime = clock.start_datetime + relativedelta(
            years=1)

        # Remove all the transactions
        del self.gl.transactions[:]

        for c in self.components:
            c.prepare_to_run(clock, period_count)

        self.negative_income_tax_total = 0

    def run(self, clock):
        """
        Execute the entity at the current clock cycle.

        :param clock: The clock containing the current execution time and
          period information.
        """

        if clock.timestep_ix >= self.period_count:
            return

        for c in self.components:
            c.run(clock, self.gl)

        self._perform_year_end_procedure(clock)

    def _perform_year_end_procedure(self, clock):
        if clock.get_datetime() >= self._curr_year_end_datetime\
         or clock.timestep_ix == self.period_count:
            gls = self.gl.structure
            year_start_date = self._prev_year_end_datetime
            year_end_date = self._curr_year_end_datetime + \
                timedelta(seconds=-1)

            self._prev_year_end_datetime = self._curr_year_end_datetime
            self._curr_year_end_datetime += relativedelta(years=1)
            if self._curr_year_end_datetime > self._exec_year_end_datetime:
                self._curr_year_end_datetime = self._exec_year_end_datetime

            gross_profit_write_off_accs = OrderedDict()
            inc_summary_write_off_accs = OrderedDict()
            sales_accs = gls.get_account_decendants(gls._acci_sales_)
            cost_of_sales_accs = gls.get_account_decendants(gls._acci_cos_)
            year_taxes = []
            for tx in self.gl.transactions:
                if tx.tx_date >= year_start_date and\
                 tx.tx_date <= year_end_date:
                    year_taxes.append(tx)
            for tx in year_taxes:
                cr_acc = gls.get_account(tx.cr_account)
                dt_acc = gls.get_account(tx.dt_account)

                if cr_acc in sales_accs:
                    if cr_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[cr_acc] += tx.amount
                    else:
                        gross_profit_write_off_accs[cr_acc] = tx.amount
                elif dt_acc in sales_accs:
                    if dt_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[dt_acc] -= tx.amount
                    else:
                        gross_profit_write_off_accs[dt_acc] = -tx.amount

                elif cr_acc in cost_of_sales_accs:
                    if cr_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[cr_acc] += tx.amount
                    else:
                        gross_profit_write_off_accs[cr_acc] = tx.amount
                elif dt_acc in cost_of_sales_accs:
                    if dt_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[dt_acc] -= tx.amount
                    else:
                        gross_profit_write_off_accs[dt_acc] = -tx.amount

                elif cr_acc.account_type == AccountType.revenue:
                    if cr_acc in inc_summary_write_off_accs:
                        inc_summary_write_off_accs[cr_acc] += tx.amount
                    else:
                        inc_summary_write_off_accs[cr_acc] = tx.amount
                elif dt_acc.account_type == AccountType.revenue:
                    if dt_acc.account_type == AccountType.revenue:
                        inc_summary_write_off_accs[dt_acc] -= tx.amount
                    else:
                        inc_summary_write_off_accs[dt_acc] = -tx.amount

                elif cr_acc.account_type == AccountType.expense:
                    if cr_acc in inc_summary_write_off_accs:
                        inc_summary_write_off_accs[cr_acc] -= tx.amount
                    else:
                        inc_summary_write_off_accs[cr_acc] = -tx.amount
                elif dt_acc.account_type == AccountType.expense:
                    if dt_acc in inc_summary_write_off_accs:
                        inc_summary_write_off_accs[dt_acc] += tx.amount
                    else:
                        inc_summary_write_off_accs[dt_acc] = tx.amount

            inc_sum = self._perform_year_end_gross_profit_and_income_summary(
                year_end_date,
                gross_profit_write_off_accs,
                inc_summary_write_off_accs)
            inc_sum = self._perform_year_end_income_tax(
                year_end_date,
                inc_sum)
            self._perform_year_end_retained_earnings(
                year_end_date,
                inc_sum)

    def _perform_year_end_gross_profit_and_income_summary(
            self,
            year_end_datetime,
            gross_profit_write_off_accounts,
            income_summary_write_off_accounts):
        gls = self.gl.structure
        gross_profit = self._perform_year_end_gross_profit(
            year_end_datetime,
            gross_profit_write_off_accounts)
        income_summary_amount = self._perform_year_end_income_summary(
            year_end_datetime,
            gross_profit,
            income_summary_write_off_accounts)

        def create_tx(dt_account, cr_account):
            tx_name = "Settle the '" + gls._acci_gross_prof_.path +\
             "' Account"
            new_tx = self.gl.create_transaction(
                tx_name,
                description=tx_name,
                tx_date=year_end_datetime,
                dt_account=dt_account,
                cr_account=cr_account,
                source=self.path,
                amount=abs(gross_profit))
            new_tx.is_closing_cr_account = True

        if gross_profit > 0:
            create_tx(gls._acci_gross_prof_.path,
                      gls._acci_inc_sum_.path)
        elif gross_profit < 0:
            create_tx(gls._acci_inc_sum_.path,
                      gls._acci_gross_prof_.path)

        return income_summary_amount

    def _perform_year_end_gross_profit(self,
                                       year_end_datetime,
                                       gross_profit_write_off_accounts):
        gls = self.gl.structure
        gross_profit = 0

        for acc, amount in gross_profit_write_off_accounts.items():
            def create_tx(dt_account, cr_account):
                tx_name = "Settle the '" + acc.path + "' Account"
                new_tx = self.gl.create_transaction(
                    tx_name,
                    description=tx_name,
                    tx_date=year_end_datetime,
                    dt_account=dt_account,
                    cr_account=cr_account,
                    source=self.path,
                    amount=abs(amount))
                new_tx.is_closing_cr_account = True

            gross_profit += amount
            if amount > 0:
                create_tx(gls._acci_gross_prof_.path, acc.path)
            elif amount < 0:
                create_tx(acc.path, gls._acci_gross_prof_.path)
        return gross_profit

    def _perform_year_end_income_summary(self,
                                         year_end_datetime,
                                         gross_profit,
                                         income_summary_write_off_accounts):
        gls = self.gl.structure
        income_summary_amount = gross_profit

        for acc, amount in income_summary_write_off_accounts.items():
            def create_tx(dt_account, cr_account):
                tx_name = "Settle the '" + acc.path + "' Account"
                new_tx = self.gl.create_transaction(
                    tx_name,
                    description=tx_name,
                    tx_date=year_end_datetime,
                    dt_account=dt_account,
                    cr_account=cr_account,
                    source=self.path,
                    amount=abs(amount))
                new_tx.is_closing_cr_account = True

            if amount > 0:
                if acc.account_type == AccountType.expense:
                    income_summary_amount -= amount
                    create_tx(gls._acci_inc_sum_.path, acc.path)
                else:
                    income_summary_amount += amount
                    create_tx(acc.path, gls._acci_inc_sum_.path)
            elif amount < 0:
                if acc.account_type == AccountType.expense:
                    income_summary_amount -= amount
                    create_tx(acc.path, gls._acci_inc_sum_.path)
                else:
                    income_summary_amount += amount
                    create_tx(gls._acci_inc_sum_.path, acc.path)
        return income_summary_amount

    def _perform_year_end_income_tax(self,
                                     year_end_datetime,
                                     income_summary_amount):
        return income_summary_amount

    def _perform_year_end_retained_earnings(self,
                                            year_end_datetime,
                                            income_summary_amount):
        def create_tx(dt_account, cr_account):
            tx_name = "Settle the '" + gls._acci_inc_sum_.path +\
                      "' Account, adjust Retained Earnings accordingly."
            new_tx = self.gl.create_transaction(
                tx_name,
                description=tx_name,
                tx_date=year_end_datetime,
                dt_account=dt_account,
                cr_account=cr_account,
                source=self.path,
                amount=income_summary_amount)
            new_tx.is_closing_cr_account = True
        gls = self.gl.structure

        if income_summary_amount > 0:
            create_tx(gls._acci_inc_sum_.path,
                      gls._accb_ret_earnings_acc_.path)
        elif income_summary_amount < 0:
            create_tx(gls._accb_ret_earnings_acc_.path,
                      gls._acci_inc_sum_.path)


if __name__ == "__main__":
    import unittest
    from auxi.modelling.business.structure_test import ActivityUnitTester
    from auxi.modelling.business.structure_test import ComponentUnitTester
    from auxi.modelling.business.structure_test import EntityUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
The purpose of this package is to provide common classes and functions for the
auxi.modelling namespace
"""


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'



#!/usr/bin/env python3
"""
This module provides testing code for the
auxi.modelling.business.basic module.
"""

import unittest
from datetime import datetime
from dateutil import relativedelta

from auxi.core.time import Clock
from auxi.modelling.business.basic import BasicActivity
from auxi.modelling.business.basic import BasicLoanActivity
from auxi.modelling.financial.des import GeneralLedger
from auxi.modelling.financial.des import GeneralLedgerStructure

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class BasicActivityUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.business.basic.BasicActivity class.
    """

    def setUp(self):
        self.object = BasicActivity("NameA",
                                    description="DescriptionA",
                                    dt_account="Bank",
                                    cr_account="Sales",
                                    amount=5000,
                                    start=datetime(2016, 2, 1),
                                    end=datetime(2017, 2, 1),
                                    interval=3)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.start_datetime, datetime(2016, 2, 1))
        self.assertEqual(self.object.end_datetime, datetime(2017, 2, 1))
        self.assertEqual(self.object.interval, 3)
        self.assertEqual(self.object.amount, 5000)
        self.assertEqual(self.object.dt_account, "Bank")
        self.assertEqual(self.object.cr_account, "Sales")

    def test__meet_exection_criteria(self):
        """
        Test that the activity only meets the execution criteria when
        it's amount is greater than 0.
        """

        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 13)
        self.assertEqual(self.object._meet_execution_criteria(5), True)
        self.object.amount = 0
        self.assertEqual(self.object._meet_execution_criteria(5), False)

    def test_run(self):
        """
        Test that the activity run method creates a transaction with an amount
        of 5000.
        """

        structure = GeneralLedgerStructure("NameA", description="DescriptionA")
        gl = GeneralLedger("NameA", structure, description="DescriptionA")
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        clock.tick()
        clock.tick()
        self.object.prepare_to_run(clock, 20)
        self.object.run(clock, gl)
        self.assertEqual(len(gl.transactions), 1)
        self.assertEqual(gl.transactions[0].amount, 5000)

    def test_get_referenced_accounts(self):
        """
        Test that the activity run method get_referenced_accounts accounts
        matches the debit and credit accounts self.object was initialised with.
        """

        result = self.object.get_referenced_accounts()
        self.assertEqual(self.object.dt_account, result[0])
        self.assertEqual(self.object.cr_account, result[1])


class BasicLoanActivityUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.business.basic.BasicLoanActivity class.
    """

    def setUp(self):
        self.gl_structure = GeneralLedgerStructure(
            "GL Structure",
            description="General Ledger Structure")
        self.gl_structure["Account Payable"].create_account(
            "Capital Loan",
            "0000")
        self.gl_structure["Expense"].create_account(
            "Interest Expense",
            "0000")

        self.gl = GeneralLedger(
            "GL",
            self.gl_structure,
            description="General Ledger")

        self.clock = Clock("NameA", start_datetime=datetime(2016, 2, 1))

        self.object = BasicLoanActivity(
            "Capital Loan",
            bank_account="Bank/Default",
            loan_account="Account Payable/Capital Loan",
            interest_account="Expense/Interest Expense",
            amount=180000,
            interest_rate=0.15,
            start=datetime(2016, 2, 1),
            duration=36,
            interval=1,
            description="Loan for Capital")

    def test_constructor(self):
        self.assertEqual(self.object.name, "Capital Loan")
        self.assertEqual(self.object.bank_account, "Bank/Default")
        self.assertEqual(
            self.object.loan_account, "Account Payable/Capital Loan")
        self.assertEqual(
            self.object.interest_account, "Expense/Interest Expense")
        self.assertEqual(self.object.amount, 180000)
        self.assertEqual(self.object.interest_rate, 0.15)
        self.assertEqual(self.object.start_datetime, datetime(2016, 2, 1))
        self.assertEqual(self.object.end_datetime, datetime(2019, 3, 1))
        self.assertEqual(self.object.duration, 36)
        self.assertEqual(self.object.interval, 1)
        self.assertEqual(self.object.description, "Loan for Capital")

    def test__meet_exection_criteria(self):
        """
        Test that the activity only meets the execution criteria when
        it's amount is greater than 0 and its duration is greater than 0.
        """

        self.object.prepare_to_run(self.clock, 13)
        self.assertEqual(self.object._meet_execution_criteria(5), True)
        self.object.amount = 0
        self.assertEqual(self.object._meet_execution_criteria(5), False)
        self.object.amount = 180000
        self.object.duration = 0
        self.assertEqual(self.object._meet_execution_criteria(5), False)

    def test_run_first_month(self):
        """
        Test that the activity run method creates a transaction with an amount
        of 180000 on the first month. No other transactions
        """

        self.object.prepare_to_run(self.clock, 1)
        self.object.run(self.clock, self.gl)
        self.assertEqual(len(self.gl.transactions), 1)
        self.assertEqual(self.gl.transactions[0].amount, 180000)

    def test_run_third_month(self):
        """
        Test that the activity run method accrued the interest correctly.
        """

        self.clock.tick()
        self.object.prepare_to_run(self.clock, 61)
        self.clock.tick()
        self.object.run(self.clock, self.gl)
        self.clock.tick()
        self.object.run(self.clock, self.gl)
        self.clock.tick()
        self.object.run(self.clock, self.gl)
        self.assertEqual(len(self.gl.transactions), 5)
        # Test the intreset rate for the first month
        # loan amount = 180000
        # First month's interest should be (180000 * 0.15) / 12 = 2250
        self.assertEqual(self.gl.transactions[1].amount, 2250)
        # Principle payment is 6239.76 (see below)
        # The interest payed the previous month is 2250
        # Making the total amount payed to the loan 6239.76 - 2250 = 3,989.76
        # New amount to calculate interest from: 180000 - 3,989.76 = 176,010.24
        # 2nd month's interest should be 176,010.24 * 0.15) / 12 = 2,200.13
        self.assertEqual("%.2f" % self.gl.transactions[3].amount, "2200.13")
        # Test the monthly principle payment
        # interest rate = 0.15: per month 0.0125.
        # loan amount = 180000
        # duration of the loan = 36
        # payment = (180000*0.0125) / (1-(1/pow((1+0.0125), 36))) = 6239.76
        self.assertEqual("%.2f" % self.gl.transactions[2].amount, "6239.76")
        self.assertEqual("%.2f" % self.gl.transactions[4].amount, "6239.76")

    def test_run_last_month(self):
        """
        Test that the activity run method creates settled the loan on the
        loans' last month and that the transactions
        """

        self.clock.tick()
        self.object.prepare_to_run(self.clock, 61)
        for i in range(0, 60):
            self.object.run(self.clock, self.gl)
            self.clock.tick()
        # Test the number of transaction that was created during the
        # loan's lifetime.
        # 2 transactions per months should be created + the initial loan amount
        # transaction. 36*2 + 1 = 73
        last_tx_ix = len(self.gl.transactions)
        self.assertEqual(len(self.gl.transactions), 73)
        # The transactions should have run only run for 36 months
        # (the duration of the loan)
        r = relativedelta.relativedelta(
            self.gl.transactions[last_tx_ix-1].tx_date,
            self.gl.transactions[0].tx_date)
        self.assertEqual(r.years * 12 + r.months, 36)

    def test_get_referenced_accounts(self):
        """
        Test that the activity run method get_referenced_accounts accounts
        matches the debit and credit accounts self.object was initialised with.
        """

        result = self.object.get_referenced_accounts()
        self.assertEqual(self.object.bank_account, result[0])
        self.assertEqual(self.object.loan_account, result[1])
        self.assertEqual(self.object.interest_account, result[2])


if __name__ == '__main__':
    unittest.main()
